# -*- coding: utf-8 -*-
"""
This module contains the user interface function and class to manage
 material properties that allow definition offine grillage elements.
 The module also contains methods that wraps ```OpenSeesPy``` material object creation.
"""

import json


def create_material(**kwargs):
    """
    User interface function to create :class:`Material` object

    The constructor of :class:`Material` object takes in three types of keyword arguments:

    #. Keyword for looking up the *ospgrillage* material library i.e. mat_lib.json.
    #. General material properties - such as E, and G
    #. Specific arguments of ``OpenSeesPy`` material library.


    :parameter:

    * code (`str`): name string of code according to mat_lib.json
    * type (`str`): Either "concrete" or "steel"
    * grade(`str`): Grade of material according to code

    The following keywords are examples of general material properties:

    :keyword:

    * E (`float`): Elastic modulus
    * G (`float`): Shear modulus

    """
    return Material(**kwargs)


class Material:
    """
    This class stores information and provides methods to parse input material properties into ``OpenSeesPy`` Material
    commands. ``OpenSeesPy`` has two types of material objects, namely UniaxialMaterial and NDMaterial.

    `Here <https://openseespydoc.readthedocs.io/en/latest/src/uniaxialMaterial.html>`_ are the information about OpenSees
    Material objects.

    As *ospgrillage* is mainly intended for bridge decks , concrete and steel makes up the primary
    materials. In turn, UniAxialMaterial object of ``OpenSeesPy`` is wrapped and used by Material class since it contains
    options for Concrete and Steel.

    The Material class also allow users to create codified material properties (e.g. AS5100).
    These material properties are stored in a material library file (mat_lib.json) for which is to the users to pass in
    the appropriate keyword arguments to create the code defined materials.

    Additionally, Material class provides method to create the mat_lib.json file, if file is not found in directory.

    .. note::

        Current version of mat_lib.json is 0.1.0
    """

    def __init__(self, **kwargs):
        """
        The constructor of Material takes in three types of keyword arguments:

        #. Keyword for looking up the *ospgrillage* material library i.e. mat_lib.json.
        #. General material properties - such as E, and G
        #. Material arguments of ```OpenSeesPy```. E.g. Opensees's Steel01 material takes isotropic hardening parameters a1
           to a4.

        The following keywords are for item (1):

        :keyword:

        * code (`str`): name string of code according to mat_lib.json
        * type (`str`): Either "concrete" or "steel"
        * grade(`str`): Grade of material according to code

        The following keywords are examples of general material properties:

        :keyword:

        * E (`float`): Elastic modulus
        * G (`float`): Shear modulus


        For developers wishing to add more material properties:

        #. if the material is a codified material, modify mat_lib.json file by adding the material following its json format.
        #. if the material is a ``OpenSees`` Material that wasn't added previously, first add its properties under :func:`get_mat_args`
            function. Then check the commands in :func:`_write_material` of :class:`OspGrillage` class.

        """
        # Instantiate variables
        self.mat_type = None  # this is the os convention for materials e.g. Concrete01
        self.op_mat_arg = None  # arguments according to OpenSeesPy
        self.units = None  # default SI units
        # assigns variables for all kwargs for codified material types
        self.code = kwargs.get("code", None)
        self.material_type = kwargs.get("material", None)
        self.material_grade = kwargs.get("grade", None)
        # assign generic material properties
        self.elastic_modulus = kwargs.get("E", None)
        self.shear_modulus = kwargs.get("G", None)
        self.poisson_ratio = kwargs.get("v", None)
        self.density = kwargs.get("rho", None)

        # properties for Concrete - symbols according to OpenSees uniaxialMaterial/Concrete
        self.fpc = kwargs.get("fpc", None)
        self.epsc0 = kwargs.get("epsc0", None)
        self.fpcu = kwargs.get("fpcu", None)
        self.epsU = kwargs.get("epsU", None)

        # properties for Steel - symbols according to OpenSees uniaxialMaterial/Steel
        self.Fy = kwargs.get("Fy", None)
        self.E0 = kwargs.get("E0", None)
        self.b = kwargs.get("b", None)  # strain -hardening ratio.
        self.a1 = kwargs.get("a1", None)
        self.a2 = kwargs.get(
            "a2", None
        )  # isotropic hardening parameter , see Ops Steel01
        self.a3 = kwargs.get("a3", None)
        self.a4 = kwargs.get("a4", None)

        # get mat lib file
        self._mat_lib = self._read_mat_lib()
        # process all inputs into relevant inputs for OpenSees material command
        self.parse_material_command()

    def parse_material_command(self):
        """
        Function to parse the material inputs into OpenSeesPy commands - this function is handled by
        :class:`ospgrillage.osp_grillage.OspGrillage`.
        """
        # check if code material is selected, if yes read from material library json
        if self.code:
            self.poisson_ratio = self._mat_lib[self.material_type][self.code][
                self.material_grade
            ]["v"]
            self.elastic_modulus = self._mat_lib[self.material_type][self.code][
                self.material_grade
            ]["E"]
            self.fpc = self._mat_lib[self.material_type][self.code][
                self.material_grade
            ]["fc"]
            self.density = self._mat_lib[self.material_type][self.code][
                self.material_grade
            ]["rho"]
            self.units = self._mat_lib[self.material_type][self.code]["units"]
        else:  # a custom material
            pass

        # perform conversion for units
        if self.units == "SI":
            self.elastic_modulus = self.elastic_modulus * 1e9
            self.fpc = self.fpc * 1e6

        # if G not defined, calculate using formula E/(2(1+v))
        if self.shear_modulus is None:
            self.shear_modulus = self.elastic_modulus / (2 * (1 + self.poisson_ratio))

        if self.material_type == "concrete":
            self.mat_type = (
                "Concrete01"  # default opensees material type to represent concrete
            )
        elif self.material_type == "steel":
            self.mat_type = (
                "Steel01"  # default opensees material type to represent steel
            )

    def get_material_args(self):
        """
        Function to get material arguments. This function is handled by
        :class:`ospgrillage.osp_grillage.OspGrillage`.
        """
        if self.mat_type == "Concrete01":
            self.op_mat_arg = [self.fpc, self.epsc0, self.fpcu, self.epsU]
        elif self.mat_type == "Steel01":
            self.op_mat_arg = [
                self.Fy,
                self.E0,
                self.b,
                self.a1,
                self.a2,
                self.a3,
                self.a4,
            ]

        # TO ADD for MORE materials

        # check if None in entries
        if None in self.op_mat_arg:
            raise Exception(
                "One or more missing/non-numeric parameters for Material: {} ".format(
                    self.mat_type
                )
            )
        return self.mat_type, self.op_mat_arg

    @staticmethod
    def _create_default_dict():
        """
        Function to create the default mat_lib.js file. The default version is 0.0.1.
        Just to make sure the JSON file is formatted correctly
        Note: 1 ksi = 6.89475728 MPa
        """

        mat_lib = {
            "concrete": {
                "AS5100-2017": {
                    "units": "SI",
                    "25MPa": {"fc": 25, "E": 26.7, "v": 0.2, "rho": 2.4e3},
                    "32MPa": {"fc": 32, "E": 30.1, "v": 0.2, "rho": 2.4e3},
                    "40MPa": {"fc": 40, "E": 32.8, "v": 0.2, "rho": 2.4e3},
                    "50MPa": {"fc": 50, "E": 34.8, "v": 0.2, "rho": 2.4e3},
                    "65MPa": {"fc": 65, "E": 37.4, "v": 0.2, "rho": 2.4e3},
                    "80MPa": {"fc": 80, "E": 39.6, "v": 0.2, "rho": 2.4e3},
                    "100MPa": {"fc": 100, "E": 42.2, "v": 0.2, "rho": 2.4e3},
                },
                "AASHTO-LRFD-8th": {
                    "units": "SI",
                    "2.4ksi": {"fc": 16.55, "E": 23.2223, "v": 0.2, "rho": 2.4027e3},
                    "3.0ksi": {"fc": 20.68, "E": 24.997, "v": 0.2, "rho": 2.4027e3},
                    "3.6ksi": {"fc": 24.82, "E": 26.547, "v": 0.2, "rho": 2.4027e3},
                    "4.0ksi": {"fc": 27.58, "E": 27.486, "v": 0.2, "rho": 2.4027e3},
                    "5.0ksi": {"fc": 34.47, "E": 29.587, "v": 0.2, "rho": 2.4027e3},
                    "6.0ksi": {"fc": 41.37, "E": 31.856, "v": 0.2, "rho": 2.4027e3},
                    "7.5ksi": {"fc": 51.71, "E": 34.999, "v": 0.2, "rho": 2.4027e3},
                    "10.0ksi": {"fc": 68.95, "E": 39.8, "v": 0.2, "rho": 2.4027e3},
                    "15.0ksi": {"fc": 103.42, "E": 48.582, "v": 0.2, "rho": 2.4027e3},
                },
            },
            "steel": {
                "AS5100.6-2004": {
                    "units": "SI",
                    "R250N": {"fy": 250, "E": 200.0, "v": 0.25, "rho": 7850},
                    "D500N": {"fy": 500, "E": 200.0, "v": 0.25, "rho": 7850},
                    "D500L": {"fy": 500, "E": 200.0, "v": 0.25, "rho": 7850},
                },
                "AASHTO-LRFD-8th": {
                    "units": "SI",
                    "A615-40": {"fy": 275.8, "E": 200.0, "v": 0.3, "rho": 7849},
                    "A615-60": {"fy": 413.67, "E": 200.0, "v": 0.3, "rho": 7849},
                    "A615-75": {"fy": 517.12, "E": 200.0, "v": 0.3, "rho": 7849},
                    "A615-80": {"fy": 551.58, "E": 200.0, "v": 0.3, "rho": 7849},
                    "A615-100": {"fy": 689.48, "E": 200.0, "v": 0.3, "rho": 7849},
                },
            },
        }

        return mat_lib

    @staticmethod
    def _write_mat_lib(mat_lib):
        """
        Write out the passed material dict
        Not to be used in the ordinary course of events at risk of overwriting
        manual edits to the mat_lib
        Used for initial creation of the mat_lib
        """

        with open("mat_lib.json", "w") as f:
            json.dump(mat_lib, f, indent=4)

    def _read_mat_lib(self):
        mat_lib = {}
        try:
            with open("mat_lib.json", "r") as f:
                mat_lib = json.load(f)
        except (FileNotFoundError, IOError):
            print("Material library unable to be read\nUsing default library")
            mat_lib = self._create_default_dict()
            self._write_mat_lib(mat_lib)
        return mat_lib

    def get_ops_material_command(self, material_tag):
        """
        Function to get ops material command
        :param material_tag: tag of material
        :type material_tag: int
        :return: str of ops material command
        """
        # e.g. concrete01 or steel01
        mat_str = None
        if self.mat_type == "Concrete01" or self.mat_type == "Steel01":
            mat_str = 'ops.uniaxialMaterial("{type}", {tag}, *{vec})\n'.format(
                type=self.mat_type, tag=material_tag, vec=self.op_mat_arg
            )
        return mat_str



# -*- coding: utf-8 -*-
"""
This module manages user interface functions and class definitions for the
grillage members. In our terminology, we define a section as the geometrical
properties of the structural elements, and the member as the combination of
the section and material properties.
"""


def create_section(**kwargs):
    """
    User interface to create :class:`Section` object.

    The constructor :class:`Section` takes the following arguments.

    :param op_ele_type: OpenSees element type - default is "elasticBeamColumn"
    :type op_ele_type: str
    :param op_section_type: OpenSees section type - default is "Elastic"
    :type op_section_type: str
    :param unit_width: Flag for unit width properties
    :type unit_width: bool

    The constructor also takes in ``kwargs`` which is summarized as follows:

    :keyword:

    * A (`float`): Cross sectional area
    * Iz (`float`): Moment of inertia about local z axis
    * Iy (`float`): Moment of inertia about local y axis
    * J (`float`): Torsional inertia - about local x axis
    * Az (`float`): Cross sectional area in the local z direction
    * Ay (`float`): Cross sectional area in the local y direction


    :return: :class:`Section` object
    """
    return Section(**kwargs)


def create_member(**kwargs):
    """
    User interface to create :class:`GrillageMember` object.

    A grillage member requires a :class:`~ospgrillage.material.Material` and a
    :class:`~ospgrillage.members.Section`

    :param section: Section class object
    :type section: :class:`Section`
    :param material: Material class object
    :type material: :class:`~ospgrillage.material.Material`
    :param member_name: Name of the grillage member (Optional)
    :type member_name: str

    :returns: :class:`~ospgrillage.members.GrillageMember` object
    """
    return GrillageMember(**kwargs)


class Section:
    """
    Class for sections. Stores geometric properties of cross sections. These information parsed into relevant
    ```OpenSeesPy``` command, creating sections in OpenSees framework.

    This class does have methods which wraps ```OpenSeesPy``` Section() commands - this is handled by
    the GrillageMember class instead.

    For developers wishing to expand the library of sections, introduce under the constructor init:

    #. The name for the new sections.
    #. The keyword arguments for the new section plus any OpenSeesPy Section keyword which is tied to.

    """

    def __init__(
        self,
        op_ele_type="elasticBeamColumn",
        mass=0,
        c_mass_flag=False,
        unit_width=False,
        op_section_type="Elastic",
        **kwargs
    ):
        """
        The constructor takes in two types of keyword arguments.

        #. General section properties - such as A, I, J for example. These properties are parses into the appropriate
           OpenSees section arguments.
        #. ```OpenSeespy``` section arguments - i.e. specific keyword for a specific ops.section() type.

        `Section <https://openseespydoc.readthedocs.io/en/latest/src/section.html>`_  information of OpenSeesPy

        Constructor takes following inputs to be parse into arguments for ```OpenSeesPy``` Sections.

        :param op_ele_type: OpenSees element type - default elasticBeamColumn
        :type op_ele_type: str
        :param op_section_type: OpenSees section type - default Elastic
        :type op_section_type: str
        :param unit_width: Flag for unit width properties
        :type unit_width: bool

        Constructor also takes the follwoing keyword arguments for defining a section.

        :keyword:
        * A (``float``): Cross sectional area
        * Iz (``float``): Moment of inertia about local z axis
        * Iy (``float``): Moment of inertia about local y axis
        * J (``float``): Torsional inertia - about local x axis
        * Az (``float``): Cross sectional area in the local z direction
        * Ay (``float``): Cross sectional area in the local y direction

        """
        # OpenSees py section type
        self.op_section_type = (
            op_section_type  # section tag based on OpenSeesPy - default is elastic
        )
        self.section_command_flag = False  # flag for parsing , check if section() command is needed. default False
        # OpenSeesPy section arguments
        # standard geometric properties
        self.A = kwargs.get("A", None)
        self.Iz = kwargs.get("Iz", None)
        self.Iy = kwargs.get("Iy", None)
        # for 3D sections
        self.Ay = kwargs.get("Ay", None)
        self.Az = kwargs.get("Az", None)
        self.J = kwargs.get("J", None)
        # for non-linear properties
        self.alpha_y = kwargs.get("alpha_y", None)
        self.alpha_z = kwargs.get("alpha_z", None)
        self.K11 = kwargs.get("K11", None)
        self.K33 = kwargs.get("K33", None)
        self.K44 = kwargs.get("K44", None)
        # for quad elements
        self.h_depth = kwargs.get("h", None)

        # types for element definition and unit width properties
        self.op_ele_type = op_ele_type
        self.unit_width = unit_width
        self.mass = mass
        self.c_mass_flag = c_mass_flag
        # keyword args
        self.num_int_pt = kwargs.get("num_int_pt", None)
        self.integration_type = kwargs.get("integration_type", None)
        # quad/tri element parameters
        self.thick = kwargs.get("thick", None)
        # for quad elements
        # here check and overwrite element if  h_depth is given, default to quad shell element
        if self.h_depth:
            self.op_section_type = "ElasticMembranePlateSection"
            self.op_ele_type = "ShellDKGQ"
        # warning checks
        E = kwargs.get("E", None)
        G = kwargs.get("G", None)
        if any([E, G]):
            raise ValueError(
                "E or G is provided to Section object. Hint: E and G are attributes of Material Object"
            )
        self.parse_section_properties()

    def parse_section_properties(self):
        # function to parse input properties
        # mainly, this is used to define less essential properties (e.g. Ay, Az) which user may not specify
        # but is required for the model commands
        # for developers, add here the section properties which needs to be parse for model command
        if self.Ay is None and self.A is not None:
            self.Ay = 0.5 * self.A
        if self.Az is None and self.A is not None:
            self.Az = 0.2 * self.A
        if self.Iy is None and self.Iz is not None:
            self.Iy = 0.2 * self.Iz


# ----------------------------------------------------------------------------------------------------------------


class GrillageMember:
    """
    Base class for defining a Grillage member. Provides methods to wrap ```OpenSeesPy``` Element() command.

    Some ```OpenSeesPy``` element() command takes in directly both material and section properties, while some requires
    first defining an ```OpenSeesPy``` material, or OpenSees section object. The role of this class is to parse the material
    and section information into its corresponding ```OpenSeesPy``` Element() command.

    `Here <https://openseespydoc.readthedocs.io/en/latest/src/element.html>`_ is more information about ```OpenSeesPy```
    Element definition.

    For developers wishing to expand the library of elements, introduce in this class:

    #. The ```OpenSeesPy``` section() command for the desire section - under ``get_ops_section_command()`` method
    #. The ```OpenSeesPy``` element() command for the element - under ``get_element_command_str()`` method


    """

    def __init__(
        self,
        section: Section,
        material,
        member_name="Undefined",
        quad_ele_flag=False,
        tri_ele_flag=False,
    ):
        """
        Constructor takes two input object. A Material, and Section object.

        :param section: Section class object
        :type section: :class:`~ospgrillage.members.Section`
        :param material: Material class object
        :type material: :class:`~ospgrillage.material.Material`
        :param member_name: Name of the grillage member (Optional)
        :type member_name: str
        """
        self.member_name = member_name
        self.section = section
        self.material = material
        self.quad_flag = quad_ele_flag
        self.tri_ele_flag = tri_ele_flag
        self.section_command_flag = True
        self.material_command_flag = True
        if any(
            [
                self.section.op_ele_type == "ElasticTimoshenkoBeam",
                self.section.op_ele_type == "elasticBeamColumn",
            ]
        ):
            self.material_command_flag = False
            self.section_command_flag = False  #
        elif any(
            [
                self.section.op_ele_type == "ShellMITC4",
                self.section.op_ele_type == "ShellDKGQ",
            ]
        ):
            self.material_command_flag = False

    def get_member_prop_arguments(self, width=1):
        # """
        # Function to output list of arguments for element type requiring list of arguments (with preceding asterisk).
        # specifically for ElasticTimoshenkoBeam, elasticBeamColumn - which no section and material tag are required
        # for developers, add the inputs for OpenSees elements here that does not require section and material tag
        # i.e. the element command directly takes in both geometrical and material properties as arguments
        # inputs
        # :return: str containing member properties in accordance with convention of OpenSees element type
        # """
        asterisk_input = None

        # if elastic Beam column elements, return str of section input
        if self.section.op_ele_type == "ElasticTimoshenkoBeam":
            if None in [
                self.material.elastic_modulus,
                self.material.shear_modulus,
                self.section.A,
                self.section.J,
                self.section.Iy,
                self.section.Iz,
                self.section.Ay,
                self.section.Az,
            ]:
                raise ValueError(
                    "One or more missing arguments for Section: {}".format(
                        self.section.op_ele_type
                    )
                )
            asterisk_input = "[{:.3e}, {:.3e}, {:.3e}, {:.3e}, {:.3e}, {:.3e}, {:.3e}, {:.3e}]".format(
                self.material.elastic_modulus,
                self.material.shear_modulus,
                self.section.A * width,
                self.section.J * width,
                self.section.Iy * width,
                self.section.Iz * width,
                self.section.Ay * width,
                self.section.Az * width,
            )
        elif self.section.op_ele_type == "elasticBeamColumn":  # eleColumn
            if None in [
                self.material.elastic_modulus,
                self.material.shear_modulus,
                self.section.A,
                self.section.J,
                self.section.Iy,
                self.section.Iz,
            ]:
                raise ValueError(
                    "One or more missing arguments for Section: {}".format(
                        self.section.op_ele_type
                    )
                )
            asterisk_input = "[{:.3e}, {:.3e}, {:.3e}, {:.3e}, {:.3e}, {:.3e}]".format(
                self.material.elastic_modulus,
                self.material.shear_modulus,
                self.section.A * width,
                self.section.J * width,
                self.section.Iy * width,
                self.section.Iz * width,
            )

            asterisk_input = "[{:.3e}, {:.3e}, {:.3e}, {:.3e}, {:.3e}, {:.3e}]".format(
                self.section.A * width,
                self.material.elastic_modulus,
                self.material.shear_modulus,
                self.section.J * width,
                self.section.Iy * width,
                self.section.Iz * width,
            )

        elif self.section.op_ele_type == "ModElasticBeam2d":
            if None in [
                self.section.A,
                self.material.elastic_modulus,
                self.section.Iz * width,
                self.section.K11,
                self.section.K33,
                self.section.K44,
            ]:
                raise ValueError(
                    "One or more missing arguments for Section: {}".format(
                        self.section.op_section_type
                    )
                )
            asterisk_input = "[{:.3e}, {:.3e}, {:.3e}, {:.3e}, {:.3e}, {:.3e}]".format(
                self.section.A * width,
                self.material.elastic_modulus,
                self.section.Iz * width,
                self.section.K11,
                self.section.K33,
                self.section.K44,
            )

        # TO be populated with more inputs for various element types

        return asterisk_input

    # Function to return argument, handled by OspGrillage
    def get_section_arguments(self, ele_width=1):
        section_args = None

        if self.section.op_section_type == "Elastic":
            section_args = [
                self.material.elastic_modulus,
                self.section.A,
                self.section.Iz,
                self.section.Iy,
                self.material.shear_modulus,
                self.section.J,
                self.section.alpha_y,
                self.section.alpha_z,
            ]

        return section_args

    def get_ops_section_command(self, section_tag=1, material_tag=None, ele_width=1):
        # here sort based on section type , return sec_str which will be populated by OpsGrillage
        # in general first format entry {} is section type, second {} is section tag, each specific entries are
        # explained in comments below each condition
        sec_str = None
        section_type = self.section.op_section_type
        if section_type == "Elastic":
            # section type, section tag, argument entries from self.get_asterisk_input()
            sec_arg = self.get_section_arguments(ele_width=ele_width)
            sec_str = 'ops.section("{type}", {tag}, *{arg})\n'.format(
                type=section_type, tag=section_tag, arg=repr(sec_arg)
            )
        elif section_type == "ElasticMembranePlateSection":
            # section type, section tag, E_mod, nu, h, rho
            sec_str = (
                'ops.section("{type}", {tag}, {E_mod}, {nu}, {h}, {rho})\n'.format(
                    type=section_type,
                    tag=section_tag,
                    E_mod=self.material.elastic_modulus,
                    nu=self.material.poisson_ratio,
                    h=self.section.h_depth,
                    rho=self.material.density,
                )
            )
        elif section_type == "PlateFiber":
            # section type, section tag, E_mod, nu, h, rho
            sec_str = 'ops.section("{type}", {tag}, {mat_tag}, {h})\n'.format(
                type=section_type,
                tag=section_tag,
                h=self.section.h_depth,
                mat_tag=material_tag,
            )

        return sec_str

    def get_element_command_str(
        self,
        ele_tag,
        node_tag_list,
        transf_tag=None,
        ele_width=1,
        materialtag=None,
        sectiontag=None,
    ) -> str:

        # Function called within OpsGrillage class `set_member()` function.
        #
        # For shell elements, n1 n2 n3 n4 are counter clockwise node (n1 being node in quadrant -1 , -1 ). This is checked
        # using sort_vertices function.
        #
        # Procedure to be called
        # 1) OpsGrillage assigns the material and section first, then returns the tag of material and section
        # 2) OpsGrillage calls get_element_command_str of GrillageMember, then it takes in material section and returns
        #  the element command ops.element() for the respective grillage member

        # format of each ele sublist
        # [node i, node j, ele group, ele tag, transtag]
        section_input = None
        ele_str = None
        if self.section.op_ele_type == "ElasticTimoshenkoBeam":
            section_input = self.get_member_prop_arguments(ele_width)
            ele_str = 'ops.element("{type}", {tag}, *{node_tag_list}, *{memberprop}, {transftag}, {mass})\n'.format(
                type=self.section.op_ele_type,
                tag=ele_tag,
                node_tag_list=node_tag_list,
                memberprop=section_input,
                transftag=transf_tag,
                mass=self.section.mass,
            )
        elif self.section.op_ele_type == "elasticBeamColumn":
            section_input = self.get_member_prop_arguments(ele_width)
            ele_str = 'ops.element("{type}", {tag}, *{node_tag_list}, *{memberprop}, {transftag}, {mass})\n'.format(
                type=self.section.op_ele_type,
                tag=ele_tag,
                node_tag_list=node_tag_list,
                memberprop=section_input,
                transftag=transf_tag,
                mass=self.section.mass,
            )
        elif self.section.op_ele_type == "nonlinearBeamColumn":
            ele_str = 'ops.element("{type}",{tag},*{node_tag_list},{num_int_pt},{sectag},{transftag},{mass})\n'.format(
                type=self.section.op_ele_type,
                tag=ele_tag,
                node_tag_list=node_tag_list,
                num_int_pt=self.section.num_int_pt,
                sectag=sectiontag,
                transftag=transf_tag,
                mass=self.section.mass,
            )

        # for shell element option
        elif self.section.op_ele_type == "ShellMITC4":
            ele_str = (
                'ops.element("{type}", {tag}, *{node_tag_list}, {sectag})\n'.format(
                    type=self.section.op_ele_type,
                    tag=ele_tag,
                    node_tag_list=node_tag_list,
                    sectag=sectiontag,
                )
            )

        elif self.section.op_ele_type == "ShellDKGQ":
            if len(node_tag_list) == 3:
                ele_type = "ShellDKGT"
            else:  # 4 node
                ele_type = "ShellDKGQ"
            ele_str = (
                'ops.element("{type}", {tag}, *{node_tag_list}, {sectag})\n'.format(
                    type=ele_type,
                    tag=ele_tag,
                    node_tag_list=node_tag_list,
                    sectag=sectiontag,
                )
            )

        elif self.section.op_ele_type == "ShellDKGT":
            ele_str = (
                'ops.element("{type}", {tag}, *{node_tag_list}, {sectag})\n'.format(
                    type=self.section.op_ele_type,
                    tag=ele_tag,
                    node_tag_list=node_tag_list,
                    sectag=sectiontag,
                )
            )

        # HERE TO POPULATE WITH MORE ELEMENT TYPES

        return ele_str



# -*- coding: utf-8 -*-
"""
This module contains the Mesh class which controls the meshing procedure of *ospgrillage*. The mesh class stores
information about the grillage mesh such as nodes, elements, and grids. Additionally, Mesh class provides methods to
mesh grillages, either orthogonal or oblique. Unlike other modules, the constructor of Mesh class is handled exclusively
by OspGrillage class.
"""
import math

from ospgrillage.static import *
from collections import namedtuple


def create_point(**kwargs):
    """
    User interface function to create a point named tuple
    :keyword:

    * x (`float` or `int`): x coordinate
    * y (`float` or `int`): y coordinate. Default = 0 for model plane of grillage
    * z (`float` or `int`): z coordinate

    :return: Point namedTuple
    """
    x = kwargs.get("x", None)
    y = kwargs.get("y", 0)
    z = kwargs.get("z", None)
    return Point(x, y, z)


# named tuple definition
Point = namedtuple("Point", ["x", "y", "z"])


class Mesh:
    """
    Base class for mesh class. The class holds information pertaining the mesh group such as element connectivity and nodes
    of the mesh object. Positional arguments are handled by OspGrillage class.

    .. note::

        As of version 0.1.0, the default mesh is a straight mesh with either Orthogonal and Oblique grids. For developers
        mesh module needs a new structure where the mesh class will have mesh types segregated into a base Mesh class
        and child class e.g. OrthogonalMesh(Mesh) class.

    """

    def __init__(
        self,
        long_dim,
        width,
        trans_dim,
        edge_dist_a,
        edge_dist_b,
        num_trans_beam,
        num_long_beam,
        skew_1,
        skew_2,
        ext_to_int_a,
        ext_to_int_b,
        pt1=Point(0, 0, 0),
        pt2=Point(0, 0, 0),
        pt3=None,
        element_counter=1,
        node_counter=1,
        transform_counter=0,
        global_x_grid_count=0,
        global_edge_count=0,
        mesh_origin=None,
        quad_ele=False,
        **kwargs
    ):

        # inputs from OspGrillage required to create mesh
        self.long_dim = long_dim
        self.trans_dim = trans_dim
        self.edge_width_a = edge_dist_a
        self.edge_width_b = edge_dist_b
        self.ext_to_int_a = ext_to_int_a
        self.ext_to_int_b = ext_to_int_b
        self.width = width
        self.num_trans_beam = num_trans_beam
        self.num_long_beam = num_long_beam
        # variables on angle and mesh type
        self.skew_1 = skew_1
        self.skew_2 = skew_2
        self.orthogonal = kwargs.get("orthogonal", False)
        self.beam_element_flag = kwargs.get(
            "create_beam_elements", True
        )  # bool to create beam elements between nodes
        # sweep path properties
        self.pt1 = pt1
        self.pt2 = pt2
        self.pt3 = pt3
        # counters to keep track of variables
        self.node_counter = node_counter
        self.element_counter = element_counter
        self.transform_counter = transform_counter
        self.global_x_grid_count = global_x_grid_count
        self.global_z_grid_count = None
        # edge construction line variables
        self.global_edge_count = global_edge_count
        self.edge_node_recorder = dict()  # key: node tag, val: unique tag for edge
        # Prefix/ rules variables here
        assigned_node_tag = []
        self.decimal_lim = 3  # variable for floating point arithmetic error
        self.skew_threshold = [11, 30]
        self.curve = False
        self.search_x_inc = 0.001
        # initiate list for nodes and elements
        self.long_ele = []
        self.trans_ele = []
        self.edge_span_ele = []
        self.y_elevation = 0
        self.mesh_design_line = []
        # dict for node and ele transform
        self.transform_dict = dict()  # key: vector xz, val: transform tag
        self.node_spec = (
            dict()
        )  # key: node tag, val: dict of node details - see technical notes
        # variables for curve mesh
        self.curve_center = []
        self.curve_radius = []
        # line / circle equation variables - instantiate
        self.m = 0
        self.c = 0
        self.r = 0
        self.R = 0
        self.d = None  # list of circle centre and circle radius [ c, r]
        # meshing properties
        if mesh_origin is None:
            self.mesh_origin = [0, 0, 0]
        else:
            self.mesh_origin = mesh_origin  # default origin
        # meshing variables
        self.first_connecting_region_nodes = []
        self.end_connecting_region_nodes = []
        self.sweep_nodes = []
        self.z_group_recorder = []
        # quad elements flag
        self.quad_ele = quad_ele
        self.max_grid_dim = None  #
        # rigid link parameters - to be assigned to beam elements representing longitudinal beams
        self.rigid_dist_y = kwargs.get("rigid_dist_y")
        self.rigid_dist_z = kwargs.get("rigid_dist_z")
        self.rigid_dist_x = kwargs.get("rigid_dist_x")
        # ------------------------------------------------------------------------------------------
        # Create sweep path obj
        self.sweep_path = SweepPath(self.pt1, self.pt2, self.pt3)
        (
            self.zeta,
            self.m,
            self.c,
        ) = self.sweep_path.get_sweep_line_properties()  # properties m,c,zeta angle
        # ------------------------------------------------------------------------------------------
        # check condition for orthogonal mesh
        if self.skew_1 != 0:
            self._check_skew(self.skew_1, self.zeta)
        elif self.skew_2 != 0:
            self._check_skew(self.skew_2, self.zeta)

        # check if angle between construction line and sweep path is sufficiently small - if greater meshing will
        # result in overlapping edges (between start and end edge) or edge construction line creating extra nodes on
        # opposite construction lines.
        if self.long_dim < self.width * np.tan(self.skew_1 / 180 * np.pi):
            raise ValueError(
                "insufficent length of grillage - causes one or more overlapping edge with extra longitudinal members"
                "due to skew angle at start edge- try using a smaller angle or larger long_dim"
            )
        elif self.long_dim < self.width * np.tan(self.skew_2 / 180 * np.pi):
            raise ValueError(
                "insufficent length of grillage resulted in overlapping edge with extra longitudinal members"
                "due to skew angle at end edge- try using a smaller angle or larger long_dim"
            )
        # ------------------------------------------------------------------------------------------
        # edge construction line 1
        self.start_edge_line = self.create_control_points(
            edge_ref_point=self.mesh_origin,
            width_z=self.width,
            edge_width_a=self.edge_width_a,
            edge_width_b=self.edge_width_b,
            edge_angle=self.skew_1,
            num_long_beam=self.num_long_beam,
            model_plane_y=self.y_elevation,
            ext_to_int_a=self.ext_to_int_a,
            ext_to_int_b=self.ext_to_int_b,
            **kwargs
        )

        # ------------------------------------------------------------------------------------------
        # edge construction line 2
        end_point_z = self.sweep_path.get_line_function(self.long_dim)
        self.end_edge_line = self.create_control_points(
            edge_ref_point=[self.long_dim, 0, end_point_z],
            width_z=self.width,
            edge_width_a=self.edge_width_a,
            edge_width_b=self.edge_width_b,
            edge_angle=self.skew_2,
            num_long_beam=self.num_long_beam,
            model_plane_y=self.y_elevation,
            ext_to_int_a=self.ext_to_int_a,
            ext_to_int_b=self.ext_to_int_b,
            **kwargs
        )
        # ------------------------------------------------------------------------------------------
        # Sweep nodes
        # nodes to be swept across sweep path varies based
        # sweeping_nodes variable is a list containing x,y,z coordinates which slope is dependant on slope at  ref point
        # of sweep nodes. slope of sweep nodes is always ORTHOGONAL to tangent of sweep path at intersection with ref
        # point
        self.sweeping_nodes = []
        # z coordinate of ref sweep nodes (relative to origin)
        self.noz = self.start_edge_line.noz
        if self.orthogonal:
            self.sweeping_nodes = self._rotate_sweep_nodes(
                self.zeta / 180 * np.pi
            )  # zeta in rad
        else:  # skew
            # sweep line of skew mesh == edge_construction line
            self.sweeping_nodes = self.start_edge_line.node_list
        self.nox = np.linspace(0, self.long_dim, self.num_trans_beam)
        # ------------------------------------------------------------------------------------------
        # create nodes and elements
        # if orthogonal, orthogonal mesh only be slayed onto a curve mesh, if skew mesh curved/arc line segment must be
        # false
        if self.orthogonal:
            # mesh start span construction line region
            self._orthogonal_meshing()

        elif not self.curve:  # Skew mesh + angle
            self._fixed_sweep_node_meshing()

        # run section grouping for longitudinal and transverse members
        self._identify_common_z_group()
        self._identify_member_groups()
        # model plane
        self.model_plane_z_groups = list(self.z_group_to_ele.keys())

    def create_control_points(self, **kwargs):
        # base version creating standard node points of control points - either start or end edge -
        # standard correspond to base model technique - grillage with beam element
        return EdgeControlLine(**kwargs)

    def _fixed_sweep_node_meshing(self):
        assigned_node_tag = []
        previous_node_tag = []
        for x_count, x_inc in enumerate(self.nox):
            for z_count, ref_point in enumerate(self.sweeping_nodes):
                # offset x and y in all points in ref points
                z_inc = np.round(
                    select_segment_function(
                        curve_flag=self.curve,
                        d=self.d,
                        m=self.m,
                        c=self.m,
                        x=x_inc,
                        r=self.r,
                    ),
                    self.decimal_lim,
                )
                node_coordinate = [
                    ref_point[0] + x_inc,
                    ref_point[1],
                    ref_point[2] + z_inc,
                ]
                self.node_spec.setdefault(
                    self.node_counter,
                    {
                        "tag": self.node_counter,
                        "coordinate": node_coordinate,
                        "x_group": x_count,
                        "z_group": z_count,
                    },
                )
                assigned_node_tag.append(self.node_counter)
                self.node_counter += 1
                # link transverse elements
                if z_count > 0:
                    # element list [element tag, node i, node j, x/z group]
                    if not self.beam_element_flag:
                        continue
                    tag = self._get_geo_transform_tag(
                        [assigned_node_tag[z_count - 1], assigned_node_tag[z_count]]
                    )
                    self.trans_ele.append(
                        [
                            self.element_counter,
                            assigned_node_tag[z_count - 1],
                            assigned_node_tag[z_count],
                            x_count,
                            tag,
                        ]
                    )
                    self.element_counter += 1

            # create longitudinal ele by linking assigned nodes @ current step with assigned nodes from previous step
            if x_count == 0:
                previous_node_tag = assigned_node_tag
                # record
                for nodes in previous_node_tag:
                    self.edge_node_recorder.setdefault(nodes, self.global_edge_count)
                self.global_edge_count += 1
            elif x_count > 0:
                for pre_node in previous_node_tag:
                    for cur_node in assigned_node_tag:
                        cur_z_group = self.node_spec[cur_node]["z_group"]
                        prev_z_group = self.node_spec[pre_node]["z_group"]
                        if cur_z_group == prev_z_group:
                            tag = self._get_geo_transform_tag([pre_node, cur_node])
                            self.long_ele.append(
                                [
                                    self.element_counter,
                                    pre_node,
                                    cur_node,
                                    cur_z_group,
                                    tag,
                                ]
                            )
                            self.element_counter += 1
                            break  # break assign long ele loop (cur node)
                # here updates the record for previous node tag step
                previous_node_tag = assigned_node_tag
                if x_count == len(self.nox) - 1:
                    for nodes in previous_node_tag:
                        self.edge_node_recorder.setdefault(
                            nodes, self.global_edge_count
                        )
                    self.global_edge_count += 1
            # reset counter and recorder for next loop x increment
            self.global_x_grid_count += 1
            assigned_node_tag = []

    def _orthogonal_meshing(self):
        """
        Main procedure to mesh nodes (and beam elements) of model based on orthogonal meshing
        """
        global sweep_nodes, z_group_recorder
        self.assigned_node_tag = []
        self.previous_node_tag = []
        self.sweep_path_points = []
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # first edge construction line
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        start_point_x = self.mesh_origin[0]
        start_point_z = self.mesh_origin[2]
        # if skew angle of edge line is below threshold for orthogonal, perform mesh as oblique for edge line
        if np.abs(self.skew_1 + self.zeta) < self.skew_threshold[0]:
            # if angle less than threshold, assign nodes of edge member as it is
            current_sweep_nodes = self.start_edge_line.node_list
            for (z_count_int, nodes) in enumerate(current_sweep_nodes):
                x_inc = start_point_x
                z_inc = start_point_z
                node_coordinate = [nodes[0] + x_inc, nodes[1], nodes[2] + z_inc]
                self.node_spec.setdefault(
                    self.node_counter,
                    {
                        "tag": self.node_counter,
                        "coordinate": node_coordinate,
                        "x_group": self.global_x_grid_count,
                        "z_group": z_count_int,
                    },
                )

                self.assigned_node_tag.append(self.node_counter)
                self.node_counter += 1
                # if loop assigned more than two nodes, link nodes as a transverse member
                if z_count_int > 0:
                    # run sub procedure to assign
                    # self.__assign_transverse_members(pre_node=self.assigned_node_tag[z_count_int - 1],
                    #                                  cur_node=self.assigned_node_tag[z_count_int])
                    if not self.beam_element_flag:
                        continue
                    if len(self.assigned_node_tag) >= 1:
                        self._assign_edge_trans_members(
                            self.assigned_node_tag[z_count_int - 1],
                            self.assigned_node_tag[z_count_int],
                            self.global_edge_count,
                        )
                        # get and link edge nodes from previous and current as skewed edge member
                        self.edge_node_recorder.setdefault(
                            self.assigned_node_tag[z_count_int - 1],
                            self.global_edge_count,
                        )
                        self.edge_node_recorder.setdefault(
                            self.assigned_node_tag[z_count_int], self.global_edge_count
                        )

                if len(self.assigned_node_tag) == len(self.noz):
                    self.first_connecting_region_nodes = self.assigned_node_tag
            self.global_x_grid_count += 1
            self.assigned_node_tag = []  # reset variable
            # print("Edge mesh @ start span completed")
        else:  # perform edge meshing with variable distance between transverse members by looping through all control
            # points of edgecontrolline
            # loop for each control point of edge line with sweep nodes
            for z_count, int_point in enumerate(self.start_edge_line.node_list):
                # search point on sweep path line whose normal intersects int_point.
                ref_point_x, ref_point_z = self._search_x_point(
                    int_point, start_point_x
                )
                # record points
                self.sweep_path_points.append(
                    [ref_point_x, self.y_elevation, ref_point_z]
                )
                # find m' of line between intersect int_point and ref point on sweep path
                m_prime, phi = get_slope(
                    [ref_point_x, self.y_elevation, ref_point_z], int_point
                )
                # rotate sweep line such that parallel to m' line
                # if skew is positive, algorithm may mistake first point as orthogonal 90 deg, specify initial m based
                # on zeta
                if self.skew_1 > 0:
                    angle = np.arctan(self.zeta / 180 * np.pi)
                else:
                    angle = np.pi / 2 - np.abs(phi)
                current_sweep_nodes = self._rotate_sweep_nodes(angle)
                # get z group of first node in current_sweep_nodes - for correct assignment in loop
                z_group = self.start_edge_line.get_node_group_z(int_point)
                # check angle condition, if skew + zeta (offset from plane)
                if 90 + self.skew_1 + self.zeta > 90:
                    sweep_nodes = current_sweep_nodes[z_count:]
                    z_group_recorder = list(range(z_group, len(current_sweep_nodes)))
                elif 90 + self.skew_1 + self.zeta < 90:
                    sweep_nodes = current_sweep_nodes[0 : (z_count + 1)]
                    z_group_recorder = (
                        list(range(0, z_group + 1)) if z_group != 0 else [0]
                    )

                # on each control point, loop through sweeping nodes to create nodes
                for (z_count_int, nodes) in enumerate(sweep_nodes):
                    x_inc = ref_point_x
                    z_inc = ref_point_z
                    node_coordinate = [nodes[0] + x_inc, nodes[1], nodes[2] + z_inc]
                    self.node_spec.setdefault(
                        self.node_counter,
                        {
                            "tag": self.node_counter,
                            "coordinate": node_coordinate,
                            "x_group": self.global_x_grid_count,
                            "z_group": z_group_recorder[z_count_int],
                        },
                    )

                    self.assigned_node_tag.append(self.node_counter)
                    self.node_counter += 1
                    # if loop assigned more than two nodes, link nodes as a transverse member
                    if not self.beam_element_flag:
                        continue
                    if z_count_int > 0:
                        # run sub procedure to assign
                        self._assign_transverse_members(
                            pre_node=self.assigned_node_tag[z_count_int - 1],
                            cur_node=self.assigned_node_tag[z_count_int],
                        )

                if not self.beam_element_flag:
                    continue
                # if loop is in first step, there is only one column of nodes, skip longitudinal assignment
                if z_count == 0:
                    self.previous_node_tag = self.assigned_node_tag
                if z_count > 0:
                    for pre_node in self.previous_node_tag:
                        for cur_node in self.assigned_node_tag:
                            cur_z_group = self.node_spec[cur_node]["z_group"]
                            prev_z_group = self.node_spec[pre_node]["z_group"]
                            if cur_z_group == prev_z_group:
                                self._assign_longitudinal_members(
                                    pre_node=pre_node,
                                    cur_node=cur_node,
                                    cur_z_group=cur_z_group,
                                )
                                break  # break assign long ele loop (cur node)

                    # if angle is positive (slope negative), edge nodes located at the first element of list
                    if len(self.assigned_node_tag) >= 1:
                        if 90 + self.skew_1 + self.zeta > 90:
                            self._assign_edge_trans_members(
                                self.previous_node_tag[0],
                                self.assigned_node_tag[0],
                                self.global_edge_count,
                            )
                            # get and link edge nodes from previous and current as skewed edge member
                            self.edge_node_recorder.setdefault(
                                self.previous_node_tag[0], self.global_edge_count
                            )
                            self.edge_node_recorder.setdefault(
                                self.assigned_node_tag[0], self.global_edge_count
                            )
                        elif 90 + self.skew_1 + self.zeta < 90:
                            self._assign_edge_trans_members(
                                self.previous_node_tag[-1],
                                self.assigned_node_tag[-1],
                                self.global_edge_count,
                            )
                            # get and link edge nodes from previous and current as skewed edge member
                            self.edge_node_recorder.setdefault(
                                self.previous_node_tag[-1], self.global_edge_count
                            )
                            self.edge_node_recorder.setdefault(
                                self.assigned_node_tag[-1], self.global_edge_count
                            )
                    # update recorder for previous node tag step
                    self.previous_node_tag = self.assigned_node_tag
                # update and reset recorders for next column of sweep nodes
                self.global_x_grid_count += 1
                if len(self.assigned_node_tag) == len(self.noz):
                    self.first_connecting_region_nodes = self.assigned_node_tag
                self.ortho_previous_node_column = self.assigned_node_tag
                self.assigned_node_tag = []

            print("Edge mesh @ start span completed")
        self.global_edge_count += 1
        # --------------------------------------------------------------------------------------------
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # second edge construction line
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # get end point of sweep line = point which sweep path intersects end span construction line
        end_point_x = self.long_dim
        # end_point_z = line_func(self.sweep_path.m,self.sweep_path.c,end_point_x)
        end_point_z = self.sweep_path.get_line_function(end_point_x)
        if np.abs(self.skew_2 + self.zeta) < self.skew_threshold[0]:
            # if angle less than threshold, assign nodes of edge member as it is
            current_sweep_nodes = self.end_edge_line.node_list
            for (z_count_int, nodes) in enumerate(current_sweep_nodes):
                x_inc = 0  # end_point_x
                z_inc = 0  # end_point_z
                node_coordinate = [nodes[0] + x_inc, nodes[1], nodes[2] + z_inc]
                self.node_spec.setdefault(
                    self.node_counter,
                    {
                        "tag": self.node_counter,
                        "coordinate": node_coordinate,
                        "x_group": self.global_x_grid_count,
                        "z_group": z_count_int,
                    },
                )

                self.assigned_node_tag.append(self.node_counter)
                self.node_counter += 1
                # if loop assigned more than two nodes, link nodes as a transverse member
                if z_count_int > 0:
                    # run sub procedure to assign
                    # self.__assign_transverse_members(pre_node=self.assigned_node_tag[z_count_int - 1],
                    #                                  cur_node=self.assigned_node_tag[z_count_int])
                    if not self.beam_element_flag:
                        continue
                    if len(self.assigned_node_tag) >= 1:
                        self._assign_edge_trans_members(
                            self.assigned_node_tag[z_count_int - 1],
                            self.assigned_node_tag[z_count_int],
                            self.global_edge_count,
                        )
                        # get and link edge nodes from previous and current as skewed edge member
                        self.edge_node_recorder.setdefault(
                            self.assigned_node_tag[z_count_int - 1],
                            self.global_edge_count,
                        )
                        self.edge_node_recorder.setdefault(
                            self.assigned_node_tag[z_count_int], self.global_edge_count
                        )
                # self.end_connecting_region_nodes = self.assigned_node_tag
                if len(self.assigned_node_tag) == len(self.noz):
                    self.end_connecting_region_nodes = self.assigned_node_tag
            self.global_x_grid_count += 1
        else:
            for z_count, int_point in enumerate(self.end_edge_line.node_list):
                # search point on sweep path line whose normal intersects int_point.
                ref_point_x, ref_point_z = self._search_x_point(
                    int_point, start_point_x
                )
                # record points
                self.sweep_path_points.append(
                    [ref_point_x, self.y_elevation, ref_point_z]
                )
                # find m' of line between intersect int_point and ref point on sweep path
                m_prime, phi = get_slope(
                    [ref_point_x, self.y_elevation, ref_point_z], int_point
                )
                # rotate sweep line such that parallel to m' line
                current_sweep_nodes = self._rotate_sweep_nodes(np.pi / 2 - np.abs(phi))
                # get z group of first node in current_sweep_nodes - for correct assignment in loop
                z_group = self.end_edge_line.get_node_group_z(
                    int_point
                )  # extract from class EdgeConstructionLine
                # check
                # condition
                if 90 + self.skew_2 + self.zeta > 90:
                    sweep_nodes = current_sweep_nodes[0 : (z_count + 1)]
                    z_group_recorder = (
                        list(range(0, z_group + 1)) if z_group != 0 else [0]
                    )
                elif 90 + self.skew_2 + self.zeta < 90:
                    sweep_nodes = current_sweep_nodes[z_count:]
                    z_group_recorder = list(range(z_group, len(current_sweep_nodes)))
                for (z_count_int, nodes) in enumerate(sweep_nodes):
                    x_inc = ref_point_x
                    z_inc = ref_point_z
                    node_coordinate = [nodes[0] + x_inc, nodes[1], nodes[2] + z_inc]
                    self.node_spec.setdefault(
                        self.node_counter,
                        {
                            "tag": self.node_counter,
                            "coordinate": node_coordinate,
                            "x_group": self.global_x_grid_count,
                            "z_group": z_group_recorder[z_count_int],
                        },
                    )

                    self.assigned_node_tag.append(self.node_counter)
                    self.node_counter += 1
                    if not self.beam_element_flag:
                        continue
                    # if loop assigned more than two nodes, link nodes as a transverse member
                    if z_count_int > 0:
                        # run sub procedure to assign
                        self._assign_transverse_members(
                            pre_node=self.assigned_node_tag[z_count_int - 1],
                            cur_node=self.assigned_node_tag[z_count_int],
                        )

                if not self.beam_element_flag:
                    continue

                # if loop is in first step, there is only one column of nodes, skip longitudinal assignment
                if z_count == 0:
                    self.previous_node_tag = self.assigned_node_tag
                if z_count > 0:
                    for pre_node in self.previous_node_tag:
                        for cur_node in self.assigned_node_tag:
                            cur_z_group = self.node_spec[cur_node]["z_group"]
                            prev_z_group = self.node_spec[pre_node]["z_group"]
                            if cur_z_group == prev_z_group:
                                self._assign_longitudinal_members(
                                    pre_node=pre_node,
                                    cur_node=cur_node,
                                    cur_z_group=cur_z_group,
                                )
                                break  # break assign long ele loop (cur node)

                    # if angle is positive (slope negative), edge nodes located at the first element of list
                    if len(self.assigned_node_tag) >= 1:
                        if 90 + self.skew_2 + self.zeta > 90:
                            self._assign_edge_trans_members(
                                self.previous_node_tag[-1],
                                self.assigned_node_tag[-1],
                                self.global_edge_count,
                            )
                            self.edge_node_recorder.setdefault(
                                self.previous_node_tag[-1], self.global_edge_count
                            )
                            self.edge_node_recorder.setdefault(
                                self.assigned_node_tag[-1], self.global_edge_count
                            )
                        elif 90 + self.skew_2 + self.zeta < 90:
                            self._assign_edge_trans_members(
                                self.previous_node_tag[0],
                                self.assigned_node_tag[0],
                                self.global_edge_count,
                            )
                            self.edge_node_recorder.setdefault(
                                self.previous_node_tag[0], self.global_edge_count
                            )
                            self.edge_node_recorder.setdefault(
                                self.assigned_node_tag[0], self.global_edge_count
                            )
                    # update recorder for previous node tag step
                    self.previous_node_tag = self.assigned_node_tag
                # update and reset recorders for next column of sweep nodes
                self.global_x_grid_count += 1
                if len(self.assigned_node_tag) == len(self.noz):
                    self.end_connecting_region_nodes = self.assigned_node_tag
                self.ortho_previous_node_column = self.assigned_node_tag
                self.assigned_node_tag = []
            self.global_edge_count += 1
            print("Edge mesh @ end span completed")
        # --------------------------------------------------------------------------------------------
        self.assigned_node_tag = []  # reset
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # remaining distance mesh with uniform spacing
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        x_first = self.first_connecting_region_nodes[0]
        x_second = self.end_connecting_region_nodes[0]
        # loop each point in self.nox
        cor_fir = self.node_spec[x_first]["coordinate"]
        cor_sec = self.node_spec[x_second]["coordinate"]
        # get x coordinate for uniform region
        self.uniform_region_x = np.linspace(cor_fir[0], cor_sec[0], self.num_trans_beam)

        for z_count, x in enumerate(self.uniform_region_x[1:-1]):
            # get slope, m at current point x
            z = self.sweep_path.get_line_function(x)
            # z = line_func(self.sweep_path.m, self.sweep_path.c, x)
            current_sweep_nodes = self._rotate_sweep_nodes(
                self.zeta / 180 * np.pi
            )  # rotating sweep nodes @ origin
            # if angle less than threshold, assign nodes of edge member as it is
            for (z_count_int, nodes) in enumerate(current_sweep_nodes):
                x_inc = x
                z_inc = z
                node_coordinate = [nodes[0] + x_inc, nodes[1], nodes[2] + z_inc]
                self.node_spec.setdefault(
                    self.node_counter,
                    {
                        "tag": self.node_counter,
                        "coordinate": node_coordinate,
                        "x_group": self.global_x_grid_count,
                        "z_group": z_count_int,
                    },
                )

                self.assigned_node_tag.append(self.node_counter)
                self.node_counter += 1
                if not self.beam_element_flag:
                    continue
                # if loop assigned more than two nodes, link nodes as a transverse member
                if z_count_int > 0:
                    # run sub procedure to assign
                    self._assign_transverse_members(
                        pre_node=self.assigned_node_tag[z_count_int - 1],
                        cur_node=self.assigned_node_tag[z_count_int],
                    )
            if not self.beam_element_flag:
                continue

            if z_count == 0:
                self.previous_node_tag = self.first_connecting_region_nodes
            elif z_count > 0 and z_count != len(self.uniform_region_x[1:-1]) - 1:
                pass
            for pre_node in self.previous_node_tag:
                for cur_node in self.assigned_node_tag:
                    cur_z_group = self.node_spec[cur_node]["z_group"]
                    prev_z_group = self.node_spec[pre_node]["z_group"]
                    if cur_z_group == prev_z_group:
                        self._assign_longitudinal_members(
                            pre_node=pre_node,
                            cur_node=cur_node,
                            cur_z_group=cur_z_group,
                        )
                        break  # break assign long ele loop (cur node)
            # update and reset recorders for next column of sweep nodes
            self.global_x_grid_count += 1
            # update previous node tag recorder
            if z_count != len(self.uniform_region_x[1:-1]) - 1:
                self.previous_node_tag = self.assigned_node_tag
                self.assigned_node_tag = []
            else:
                self.previous_node_tag = self.assigned_node_tag
                self.assigned_node_tag = self.end_connecting_region_nodes

        # Extra step to connect uniform region with nodes along end span edge region
        for pre_node in self.previous_node_tag:
            if not self.beam_element_flag:
                break
            for cur_node in self.assigned_node_tag:
                cur_z_group = self.node_spec[cur_node]["z_group"]
                prev_z_group = self.node_spec[pre_node]["z_group"]
                if cur_z_group == prev_z_group:
                    self._assign_longitudinal_members(
                        pre_node=pre_node, cur_node=cur_node, cur_z_group=cur_z_group
                    )
                    break

    # ------------------------------------------------------------------------------------------
    # ------------------------------------------------------------------------------------------
    def _assign_transverse_members(self, pre_node, cur_node):
        tag = self._get_geo_transform_tag([pre_node, cur_node])
        self.trans_ele.append(
            [self.element_counter, pre_node, cur_node, self.global_x_grid_count, tag]
        )
        self.element_counter += 1

    def _assign_longitudinal_members(self, pre_node, cur_node, cur_z_group):
        tag = self._get_geo_transform_tag([pre_node, cur_node])
        self.long_ele.append(
            [self.element_counter, pre_node, cur_node, cur_z_group, tag]
        )
        self.element_counter += 1

    def _assign_edge_trans_members(
        self, previous_node_tag, assigned_node_tag, edge_counter
    ):
        tag = self._get_geo_transform_tag([previous_node_tag, assigned_node_tag])
        self.edge_span_ele.append(
            [
                self.element_counter,
                previous_node_tag,
                assigned_node_tag,
                edge_counter,
                tag,
            ]
        )
        self.element_counter += 1

    # ------------------------------------------------------------------------------------------
    def _identify_common_z_group(self):
        # dict common element group to z group
        self.common_z_group_element = dict()
        self.common_z_group_element[0] = [
            0,
            len(self.noz) - 1,
        ]  # edge beams top and bottom edge

        # start with case where only 2 grid line in long dir, no exterior and interior beams
        exterior_beam_1_group = []
        exterior_beam_2_group = []
        interior_beam_group = list(
            range(2, len(self.noz) - 2)
        )  # default [] when len(self.noz) < 2
        # if more than 3 grid lines
        if len(self.noz) > 2:
            if len(self.noz) == 3:  # 3 grid lines, 2 edge and 1 interior
                interior_beam_group = [1]  # overwrite interior
            elif len(self.noz) == 4:  # 4 grid line, 2 edge exterior 1 and exterior 2
                exterior_beam_1_group = [1]
                interior_beam_group = []
                exterior_beam_2_group = [len(self.noz) - 2]
            else:
                exterior_beam_1_group = [1]
                exterior_beam_2_group = [len(self.noz) - 2]
        self.common_z_group_element[1] = exterior_beam_1_group  # exterior 1
        self.common_z_group_element[2] = interior_beam_group  # interior
        self.common_z_group_element[3] = exterior_beam_2_group  # exterior 2

    def _identify_member_groups(self):
        """
        Abstracted method handled by either orthogonal_mesh() or skew_mesh() function
        to identify member groups based on node spacings in orthogonal directions.

        :return: Set variable `group_ele_dict` according to
        """

        # dict node tag to width in z direction , and neighbouring node
        self.node_width_z_dict = dict()
        self.node_connect_z_dict = dict()
        for ele in self.long_ele:
            d1 = []  # d for distance
            d2 = []
            p1 = []
            p2 = []
            n1 = [
                trans_ele
                for trans_ele in self.trans_ele
                if trans_ele[1] == ele[1] or trans_ele[2] == ele[1]
            ]
            n2 = [
                trans_ele
                for trans_ele in self.trans_ele
                if trans_ele[1] == ele[2] or trans_ele[2] == ele[2]
            ]
            for item in n1:
                d1.append(
                    [
                        np.abs(a - b)
                        for (a, b) in zip(
                            self.node_spec[item[1]]["coordinate"],
                            self.node_spec[item[2]]["coordinate"],
                        )
                    ]
                )
                if item[1] != ele[1] and item[1] != ele[2]:
                    p1.append(item[1])
                if item[2] != ele[1] and item[2] != ele[2]:
                    p1.append(item[2])

            for item in n2:
                d2.append(
                    [
                        np.abs(a - b)
                        for (a, b) in zip(
                            self.node_spec[item[1]]["coordinate"],
                            self.node_spec[item[2]]["coordinate"],
                        )
                    ]
                )
                if item[1] != ele[1] and item[1] != ele[2]:
                    p2.append(item[1])
                if item[2] != ele[1] and item[2] != ele[2]:
                    p2.append(item[2])
            # list, [ele tag, ele width (left and right)]
            self.node_width_z_dict.setdefault(ele[1], d1)
            self.node_width_z_dict.setdefault(ele[2], d2)
            self.node_connect_z_dict.setdefault(ele[1], p1)
            self.node_connect_z_dict.setdefault(ele[2], p2)

        # dict z to long ele
        self.z_group_to_ele = dict()
        for count, node in enumerate(self.noz):
            self.z_group_to_ele[count] = [
                ele for ele in self.long_ele if ele[3] == count
            ]

        self.global_z_grid_count = max(self.z_group_to_ele.keys()) + 1
        # dict x to trans ele
        self.x_group_to_ele = dict()
        for count in range(0, self.global_x_grid_count):
            self.x_group_to_ele[count] = [
                ele for ele in self.trans_ele if ele[3] == count
            ]
        # dict edge counter to ele
        self.edge_group_to_ele = dict()
        for count in range(0, self.global_edge_count + 1):
            self.edge_group_to_ele[count] = [
                ele for ele in self.edge_span_ele if ele[3] == count
            ]
        # dict node tag to width in x direction
        self.node_width_x_dict = dict()
        self.node_connect_x_dict = dict()
        for ele in self.trans_ele:
            d1 = []
            d2 = []
            p1 = []
            p2 = []
            n1 = [
                long_ele
                for long_ele in self.long_ele
                if long_ele[1] == ele[1] or long_ele[2] == ele[1]
            ]
            n2 = [
                long_ele
                for long_ele in self.long_ele
                if long_ele[1] == ele[2] or long_ele[2] == ele[2]
            ]
            for item in n1:
                d1.append(
                    [
                        np.abs(a - b)
                        for (a, b) in zip(
                            self.node_spec[item[1]]["coordinate"],
                            self.node_spec[item[2]]["coordinate"],
                        )
                    ]
                )
                if item[1] != ele[1] and item[1] != ele[2]:
                    p1.append(item[1])
                if item[2] != ele[1] and item[2] != ele[2]:
                    p1.append(item[2])
            for item in n2:
                d2.append(
                    [
                        np.abs(a - b)
                        for (a, b) in zip(
                            self.node_spec[item[1]]["coordinate"],
                            self.node_spec[item[2]]["coordinate"],
                        )
                    ]
                )
                if item[1] != ele[1] and item[1] != ele[2]:
                    p2.append(item[1])
                if item[2] != ele[1] and item[2] != ele[2]:
                    p2.append(item[2])
            # list, [ele tag, ele width (left and right)]
            self.node_width_x_dict.setdefault(ele[1], d1)
            self.node_width_x_dict.setdefault(ele[2], d2)
            self.node_connect_x_dict.setdefault(ele[1], p1)
            self.node_connect_x_dict.setdefault(ele[2], p2)

        for ele in self.edge_span_ele:
            d1 = []
            d2 = []
            p1 = []
            p2 = []
            n1 = [
                long_ele
                for long_ele in self.long_ele
                if long_ele[1] == ele[1] or long_ele[2] == ele[1]
            ]
            n2 = [
                long_ele
                for long_ele in self.long_ele
                if long_ele[1] == ele[2] or long_ele[2] == ele[2]
            ]
            for item in n1:
                d1.append(
                    [
                        np.abs(a - b)
                        for (a, b) in zip(
                            self.node_spec[item[1]]["coordinate"],
                            self.node_spec[item[2]]["coordinate"],
                        )
                    ]
                )
                if item[1] != ele[1] and item[1] != ele[2]:
                    p1.append(item[1])
                if item[2] != ele[1] and item[2] != ele[2]:
                    p1.append(item[2])
            for item in n2:
                d2.append(
                    [
                        np.abs(a - b)
                        for (a, b) in zip(
                            self.node_spec[item[1]]["coordinate"],
                            self.node_spec[item[2]]["coordinate"],
                        )
                    ]
                )
                if item[1] != ele[1] and item[1] != ele[2]:
                    p2.append(item[1])
                if item[2] != ele[1] and item[2] != ele[2]:
                    p2.append(item[2])
            # list, [ele tag, ele width (left and right)]
            self.node_width_x_dict.setdefault(ele[1], d1)
            self.node_width_x_dict.setdefault(ele[2], d2)
            self.node_connect_x_dict.setdefault(ele[1], p1)
            self.node_connect_x_dict.setdefault(ele[2], p2)
        # create self.grid_number_dict, dict key = grid number, val = long and trans ele in grid
        self.grid_number_dict = dict()
        counter = 0
        for node_tag in self.node_spec.keys():
            # get the surrounding nodes
            x_vicinity_nodes = self.node_connect_x_dict.get(node_tag, [])
            z_vicinity_nodes = self.node_connect_z_dict.get(node_tag, [])
            for x_node in x_vicinity_nodes:
                xg = self.node_spec[x_node]["x_group"]
                for z_node in z_vicinity_nodes:
                    zg = self.node_spec[z_node]["z_group"]
                    # find the 3rd bounding node
                    n3 = [
                        n["tag"]
                        for n in self.node_spec.values()
                        if n["x_group"] == xg and n["z_group"] == zg
                    ]
                    if n3:
                        n3 = n3[0]
                        if not any(
                            [
                                node_tag in d
                                and x_node in d
                                and z_node in d
                                and n3 in d
                                for d in self.grid_number_dict.values()
                            ]
                        ):
                            self.grid_number_dict.setdefault(
                                counter, [node_tag, x_node, n3, z_node]
                            )
                            counter += 1
                    else:  # list is empty
                        if not any(
                            [
                                node_tag in d and x_node in d and z_node in d
                                for d in self.grid_number_dict.values()
                            ]
                        ):
                            self.grid_number_dict.setdefault(
                                counter, [node_tag, x_node, n3, z_node]
                            )
                            counter += 1

        # dict of grid number return vicinity grid number in a subdict {'x-1': 'x+1', 'z-1' , 'z+1'}
        self.grid_vicinity_dict = dict()
        for k, grid in self.grid_number_dict.items():
            current_x_group = []
            current_z_group = []
            current_x = []
            current_z = []

            grid_number_record = []
            if [] in grid:
                grid.remove([])
            for node in grid:
                grid_number_record += [
                    i
                    for i, x in enumerate(
                        [node in n for n in self.grid_number_dict.values()]
                    )
                    if x
                ]
                current_x_group.append(self.node_spec[node]["x_group"])
                current_z_group.append(self.node_spec[node]["z_group"])
                current_x.append(self.node_spec[node]["coordinate"][0])
                current_z.append(self.node_spec[node]["coordinate"][2])
            current_x_group = list(np.unique(current_x_group))
            current_z_group = list(np.unique(current_z_group))
            current_x = list(np.unique(current_x))
            current_z = list(np.unique(current_z))
            grid_number_record = np.unique(grid_number_record)
            # loop to characterize the grid for current
            subdict = {}
            for neighbour in grid_number_record:
                if neighbour == k:  # identical , current grid
                    continue
                x_group = []  # initialize variables
                x_coor = []
                z_group = []
                z_coor = []
                # loop each node in the vicintiy grids
                for nodes in self.grid_number_dict[neighbour]:
                    if not nodes:
                        continue
                    x_group.append(self.node_spec[nodes]["x_group"])
                    z_group.append(self.node_spec[nodes]["z_group"])
                    x_coor.append(self.node_spec[nodes]["coordinate"][0])
                    z_coor.append(self.node_spec[nodes]["coordinate"][2])
                x_group = list(np.unique(x_group))
                z_group = list(np.unique(z_group))
                x_coor = list(np.unique(x_coor))
                z_coor = list(np.unique(z_coor))
                # if x groups are identical, neighbour grid is either top or bottom of the element
                if all(a in current_x_group for a in x_group):
                    # compare z max
                    if max(z_coor) > max(current_z):
                        subdict["top"] = neighbour
                    else:
                        subdict["bottom"] = neighbour
                # if x groups are identical, neighbour grid is either left or right of the element
                if all(a in current_z_group for a in z_group):
                    if max(x_coor) > max(current_x):
                        subdict["right"] = neighbour
                    else:
                        subdict["left"] = neighbour
            self.grid_vicinity_dict.setdefault(k, subdict)

    def _get_geo_transform_tag(self, ele_nodes, offset=None):
        # offset is not used in version 0.1.0
        if offset is None:
            offset = []  #
        # function called for each element, assign
        node_i = self.node_spec[ele_nodes[0]]["coordinate"]
        node_j = self.node_spec[ele_nodes[1]]["coordinate"]
        vxz = self._get_vector_xz(node_i, node_j)
        vxz = [np.round(num, decimals=self.decimal_lim) for num in vxz]
        tag_value = self.transform_dict.setdefault(
            repr(vxz) + "|" + repr(offset), self.transform_counter + 1
        )
        if tag_value > self.transform_counter:
            self.transform_counter = tag_value
        return tag_value

    def _check_skew(self, edge_skew_angle, zeta):
        # zeta in DEGREES
        # if mesh type is beyond default allowance threshold of 11 degree and 30 degree, return exception
        if np.abs(edge_skew_angle - zeta) <= self.skew_threshold[0] and self.orthogonal:
            # return error
            raise Exception(
                "Skew angle too small for orthogonal, minimum edge skew angle for an orthogonal mesh is {}".format(
                    self.skew_threshold[0]
                )
            )
        elif (
            np.abs(edge_skew_angle - zeta) >= self.skew_threshold[1]
            and not self.orthogonal
        ):
            self.orthogonal = True
            raise Exception(
                "Skew angle too large for Oblique mesh, maximum edge skew angle for Oblique mesh is {}".format(
                    (self.skew_threshold[1])
                )
            )
            # raise Exception('Oblique mesh not allowed for angle greater than {}'.format(self.skew_threshold[1]))

    # ------------------------------------------------------------------------------------------
    @staticmethod
    def _get_vector_xz(node_i, node_j):
        """
        Encapsulated function to identify a vector parallel to the plane of local x and z axis of the element. The
        vector is required for geomTransf() command
        - see geomTransf_.

        .. _geomTransf: https://openseespydoc.readthedocs.io/en/latest/src/geomTransf.html

        """
        # Function to calculate vector xz used for geometric transformation of local section properties
        # return: vector parallel to plane xz of member (see geotransform Opensees) for skew members (member tag 5)

        # vector rotate 90 deg clockwise (x,y) -> (y,-x)
        # [breadth width] is a vector parallel to skew
        xi = node_j[0] - node_i[0]
        zi = node_j[2] - node_i[2]
        x = zi
        z = -xi
        # normalize vector
        length = np.sqrt(x ** 2 + z ** 2)
        x1 = x / length

        z1 = z / length
        return [x1, 0, z1]  # here y axis is normal to model plane

    def _rotate_sweep_nodes(self, zeta):
        sweep_nodes_x = [0] * len(
            self.noz
        )  # line is orthogonal at the start of sweeping path
        # rotate for inclination at origin
        sweep_nodes_x = [
            x * np.cos(zeta) - y * np.sin(zeta) for x, y in zip(sweep_nodes_x, self.noz)
        ]
        sweep_nodes_z = [
            y * np.cos(zeta) + x * np.sin(zeta) for x, y in zip(sweep_nodes_x, self.noz)
        ]

        sweeping_nodes = [
            [x + self.mesh_origin[0], y + self.mesh_origin[1], z + self.mesh_origin[2]]
            for x, y, z in zip(
                (sweep_nodes_x), [self.y_elevation] * len(self.noz), sweep_nodes_z
            )
        ]

        return sweeping_nodes

    def _search_x_point(self, int_point, start_point_y=0, line_function=None):
        start_point_x = int_point[0]
        min_found = False
        max_loop = 1000
        loop_counter = 0
        z0 = None
        inc = self.search_x_inc
        convergence_check = []
        bounds = []
        while not min_found:
            z0 = self.sweep_path.get_line_function(start_point_x)
            z_ub = self.sweep_path.get_line_function(start_point_x + inc)
            z_lb = self.sweep_path.get_line_function(start_point_x - inc)
            d0 = find_min_x_dist(
                [int_point], [[start_point_x, self.y_elevation, z0]]
            ).tolist()

            d_ub = find_min_x_dist(
                [int_point], [[start_point_x + inc, self.y_elevation, z_ub]]
            ).tolist()

            d_lb = find_min_x_dist(
                [int_point], [[start_point_x - inc, self.y_elevation, z_lb]]
            ).tolist()

            if d_lb > d0 and d_ub > d0:
                min_found = True
            elif d_lb < d0 and d_ub > d0:
                start_point_x = start_point_x - inc
            elif d_lb > d0 and d_ub < d0:
                start_point_x = start_point_x + inc

            loop_counter += 1
            if loop_counter > max_loop:
                break

        return start_point_x, z0


class EdgeControlLine:
    """
    Main class for edge control points. This class holds information for node meshing points of model edges - start and
    end spans. This class is used in BeamMesh class.
    """

    def __init__(
        self,
        edge_ref_point,
        width_z,
        edge_width_a,
        edge_width_b,
        edge_angle,
        num_long_beam,
        model_plane_y,
        feature="standard",
        ext_to_int_a=None,
        ext_to_int_b=None,
        **kwargs
    ):

        # distance between interior beam and exterior beam
        self.ext_to_int_a = ext_to_int_a
        self.ext_to_int_b = ext_to_int_b
        # set variables
        self.edge_ref_point = edge_ref_point
        self.width_z = width_z
        self.edge_width_a = edge_width_a
        self.edge_width_b = edge_width_b
        self.num_long_beam = num_long_beam
        self.edge_angle = edge_angle
        self.feature = feature

        # for shell
        self.z_group_master_pair_list = []
        self.node_z_pair_list_value = []
        # counter
        self.customize = kwargs.get("custom", None)  # get the list
        if self.customize and not isinstance(self.customize, list):
            raise Exception(
                "Invalid custom control point format: Hint - accepts list of float or int"
            )

        # calculations
        # TODO TEST
        # array containing z coordinate of edge construction line
        last_girder = self.width_z - self.edge_width_b  # coord of exterior main beam 2

        last_interior = (
            last_girder - self.ext_to_int_b if self.ext_to_int_b is not None else None
        )
        first_interior = (
            self.edge_width_a + self.ext_to_int_a
            if self.ext_to_int_a is not None
            else None
        )

        # check cases of customize edge control points
        if (
            not first_interior and not last_interior
        ):  # no custom dist between interior and exterior
            nox_girder = np.linspace(
                start=self.edge_width_a, stop=last_girder, num=self.num_long_beam - 2
            )
        elif first_interior and not last_interior:
            nox_girder = np.hstack(
                (
                    edge_width_a,
                    np.linspace(
                        start=first_interior,
                        stop=last_girder,
                        num=self.num_long_beam - 2 - 1,
                    ),
                )
            )
        elif not first_interior and last_interior:
            nox_girder = np.hstack(
                (
                    np.linspace(
                        start=self.edge_width_a,
                        stop=last_interior,
                        num=self.num_long_beam - 2 - 1,
                    ),
                    last_girder,
                )
            )

        else:  # both have custom interior ext distance
            nox_girder = np.linspace(
                start=first_interior, stop=last_interior, num=self.num_long_beam - 4
            )
            nox_girder = np.hstack((edge_width_a, nox_girder))
            nox_girder = np.hstack((nox_girder, last_girder))

        # create self.noz points
        self._create_trans_grid(nox_girder=nox_girder)
        # if self.feature == "standard":
        #     self.noz = np.hstack((np.hstack((0, nox_girder)), self.width_z))
        # elif self.feature == "shell_link":

        # if negative angle, create edge_node_x based on negative angle algorithm, else positive angle algorithm
        if self.edge_angle <= 0:
            edge_node_x = [
                -(z * np.tan(self.edge_angle / 180 * np.pi)) for z in self.noz
            ]  # rotate z by edge angle
            self.node_list = [
                [
                    x + self.edge_ref_point[0],
                    y + self.edge_ref_point[1],
                    z + self.edge_ref_point[2],
                ]
                for x, y, z in zip(
                    edge_node_x, [model_plane_y] * len(self.noz), self.noz
                )
            ]
        else:
            edge_node_x = [
                -(z * np.tan(self.edge_angle / 180 * np.pi)) for z in self.noz
            ]
            self.node_list = [
                [
                    x + self.edge_ref_point[0],
                    y + self.edge_ref_point[1],
                    z + self.edge_ref_point[2],
                ]
                for x, y, z in zip(
                    edge_node_x, [model_plane_y] * len(self.noz), self.noz
                )
            ]

        self.z_group = list(range(0, len(self.noz)))

        self.slope = (
            -1 / np.tan(self.edge_angle / 180 * np.pi) if self.edge_angle != 0 else None
        )
        self.c = (
            get_y_intcp(self.slope, self.edge_ref_point[0], self.edge_ref_point[2])
            if self.edge_angle != 0
            else None
        )

    def get_node_group_z(self, coordinate):
        # return list of zgroup
        group = self.node_list.index(coordinate)
        return group

    def _create_trans_grid(self, nox_girder):
        self.noz = np.hstack((np.hstack((0, nox_girder)), self.width_z))


class ShellEdgeControlLine(EdgeControlLine):
    """
    Child class of EdgeControlLine. This class is used by ShellMesh class
    """

    def __init__(
        self,
        edge_ref_point,
        width_z,
        edge_width_a,
        edge_width_b,
        edge_angle,
        num_long_beam,
        model_plane_y,
        feature="standard",
        ext_to_int_a=None,
        ext_to_int_b=None,
        **kwargs
    ):
        # get properties specific to shell mesh
        self.beam_width = kwargs.get(
            "beam_width", None
        )  # information from kwargs of Shellmodel class
        self.max_mesh_size_z = kwargs.get(
            "max_mesh_size_z"
        )  # information from kwargs of Shellmodel class
        self.max_mesh_size_x = kwargs.get(
            "max_mesh_size_x"
        )  # information from kwargs of Shellmodel class
        super().__init__(
            edge_ref_point,
            width_z,
            edge_width_a,
            edge_width_b,
            edge_angle,
            num_long_beam,
            model_plane_y,
            feature,
            ext_to_int_a,
            ext_to_int_b,
            **kwargs
        )

    # function specific to shell edge line
    def _get_shell_z_group_pair(self):
        for node_z_pair in self.node_z_pair_list_value:
            # get first and second z group of paired z group corresponding to links to offset line element
            list_index = [
                self.noz.index(node_z_pair[0]),
                self.noz.index(node_z_pair[1]),
            ]
            self.z_group_master_pair_list.append(list_index)

    def _create_trans_grid(self, nox_girder):
        z_spacing = self.beam_width / 2
        mesh_z_spacing = self.max_mesh_size_z
        self.beam_position = np.hstack(
            (np.hstack((0, nox_girder)), self.width_z)
        )  # default noz representing beam position
        shell_noz = [self.edge_ref_point[2]]  # first and last node z
        for beam_node_z in self.beam_position[1:-1]:
            local_list = (
                []
            )  # create local list of control points (local to each beam group)
            # create external control points between beam groups
            num_points_external = math.ceil(
                (beam_node_z - z_spacing - shell_noz[-1]) / mesh_z_spacing
            )
            local_list += np.linspace(
                shell_noz[-1], beam_node_z - z_spacing, num_points_external + 1
            ).tolist()
            # local_list += np.linspace(shell_noz[-1], beam_node_z - z_spacing, 2).tolist()  # external
            # create internal control points within the beam group
            num_points_internal = math.ceil(z_spacing * 2 / mesh_z_spacing)
            local_list += np.linspace(
                beam_node_z - z_spacing,
                beam_node_z + z_spacing,
                num_points_internal + 1,
            ).tolist()[1:]
            self.node_z_pair_list_value.append(
                [beam_node_z - z_spacing, beam_node_z + z_spacing]
            )
            shell_noz += local_list[1:]
        # create points between exterior beam to outer edge beam
        num_points_external = math.ceil((self.width_z - shell_noz[-1]) / mesh_z_spacing)
        shell_noz += np.linspace(
            shell_noz[-1], self.width_z, num_points_external + 1
        ).tolist()[1:]
        shell_noz.sort()
        self.noz = shell_noz  #
        self._get_shell_z_group_pair()


class SweepPath:
    """
    Main class for sweep path. Sweep path is assigned to an EdgeControlLine class in order to create mesh of nodes
    across its path. The constuctor is handled by Mesh classes ( either base or concrete classes)
    """

    def __init__(self, pt1: Point, pt2: Point, pt3: Point = None):
        """

        :param pt1: Namedtuple Point of first coordinate
        :param pt2: Namedtuple Point of second coordinate
        :param pt3: Namedtuple Point of third coordinate
        """
        self.pt1 = pt1  # default first
        self.pt2 = pt2  # default second / last for linear line
        self.pt3 = pt3  # default mid point of a curved line defined using 3 points
        self.decimal_lim = 4
        self.curve_path = False
        # return variables
        self.zeta = None
        self.m = None
        self.c = 0  # Default:0 , sweep path intersects origin

    def get_sweep_line_properties(self):
        if self.pt3 is not None:
            try:
                self.d = find_circle(
                    x1=0,
                    y1=0,
                    x2=self.pt2.x,
                    y2=self.pt2.z,
                    x3=self.pt3.x,
                    y3=self.pt3.z,
                )
                self.curve_path = True
            except ZeroDivisionError:
                return Exception(
                    "Zero div error. Point 3 not valid to construct curve line"
                )
            # procedure
            # get tangent at origin
            self.zeta = 0
            # get tangent at end of curve line (intersect with second construction line)

        else:
            # construct straight line sweep path instead
            self.d = None
            # procedure to identify straight line segment pinpointing length of grillage
            points = [(self.pt1.x, self.pt1.z), (self.pt2.x, self.pt2.z)]
            x_coords, y_coords = zip(*points)
            A = np.vstack([x_coords, np.ones(len(x_coords))]).T
            m, c = np.linalg.lstsq(A, y_coords, rcond=None)[0]
            self.m = round(m, self.decimal_lim)
            # self.c = 0  # default 0  to avoid arithmetic error
            zeta = np.arctan(
                m
            )  # initial angle of inclination of sweep line about mesh origin
            self.zeta = zeta / np.pi * 180  # rad to degrees
            self.eta_angle = self.zeta

        return self.zeta, self.m, self.c

    def get_line_function(self, x):
        if not self.curve_path:
            # straight line
            return line_func(self.m, self.c, x)
        else:
            # TODO for curve line
            pass


# ---------------------------------------------------------------------------------------------------------------------
# concrete classes of Mesh
class BeamMesh(Mesh):
    """
    Concrete class for Mesh class. This is the default Mesh class for Beam grillage
    """

    def __init__(
        self,
        long_dim,
        width,
        trans_dim,
        edge_dist_a,
        edge_dist_b,
        num_trans_beam,
        num_long_beam,
        skew_1,
        skew_2,
        ext_to_int_a,
        ext_to_int_b,
        **kwargs
    ):
        """
        Subclass for Mesh with beam. This class creates elements where:
        - nodes of longitudinal beams are offset in global y direction
        - nodes of transverse slab are offset in direction of element axial direction

        :param long_dim:
        :param width:
        :param trans_dim:
        :param edge_dist_a:
        :param edge_dist_b:
        :param num_trans_beam:
        :param num_long_beam:
        :param skew_1:
        :param skew_2:
        :param ext_to_int_a:
        :param ext_to_int_b:
        """
        # instantiate variables specific for current mesh subclass

        # super init to create mesh of grillage @ model plane = 0 using base class init
        super().__init__(
            long_dim,
            width,
            trans_dim,
            edge_dist_a,
            edge_dist_b,
            num_trans_beam,
            num_long_beam,
            skew_1,
            skew_2,
            ext_to_int_a,
            ext_to_int_b,
            **kwargs
        )


class BeamLinkMesh(Mesh):
    """
    Concrete class for Mesh class. This Mesh class is for model type 2: Beam grillage with rigid links and offsets
    """

    def __init__(
        self,
        long_dim,
        width,
        trans_dim,
        edge_dist_a,
        edge_dist_b,
        num_trans_beam,
        num_long_beam,
        skew_1,
        skew_2,
        ext_to_int_a,
        ext_to_int_b,
        **kwargs
    ):
        """
        Subclass for Mesh with beam. This class creates elements where:
        - nodes of longitudinal beams are offset in global y direction
        - nodes of transverse slab are offset in direction of element axial direction

        :param long_dim:
        :param width:
        :param trans_dim:
        :param edge_dist_a:
        :param edge_dist_b:
        :param num_trans_beam:
        :param num_long_beam:
        :param skew_1:
        :param skew_2:
        :param ext_to_int_a:
        :param ext_to_int_b:
        """
        # instantiate variables specific for beam link model
        self.beam_width = kwargs.get("beam_width", None)
        self.web_thick = kwargs.get("web_thick", None)
        self.centroid_dist_y = kwargs.get("centroid_dist_y", 0)
        if any([self.beam_width is None and self.web_thick is None]):
            self.offset_z_dist = 0
        else:
            self.offset_z_dist = self.beam_width / 2 - self.web_thick / 2

        # super init to create mesh of grillage @ model plane = 0 using base class init
        super().__init__(
            long_dim,
            width,
            trans_dim,
            edge_dist_a,
            edge_dist_b,
            num_trans_beam,
            num_long_beam,
            skew_1,
            skew_2,
            ext_to_int_a,
            ext_to_int_b,
            **kwargs
        )

    def _get_geo_transform_tag(self, ele_nodes, offset=None):
        """
        overwrite base class method to get geom transf of beam element
        """
        global_offset = []  #
        local_offset = []  #
        # function called for each element, assign
        node_i = self.node_spec[ele_nodes[0]]["coordinate"]
        node_j = self.node_spec[ele_nodes[1]]["coordinate"]
        def_l = find_min_x_dist([node_i], [node_j]).tolist()[0][
            0
        ]  # distance between i and j, returned as 2D array
        mid_node = [
            (node_i[0] + node_j[0]) / 2,
            (node_i[1] + node_j[1]) / 2,
            (node_i[2] + node_j[2]) / 2,
        ]
        vxz = self._get_vector_xz(node_i, node_j)

        # determine local offset of node based on element groups, get global offset for node i and j of geomtransf
        # if element is a longitudinal, set global y offset (for longitudinal beam)
        if (
            self.node_spec[ele_nodes[1]]["z_group"]
            == self.node_spec[ele_nodes[0]]["z_group"]
        ):
            # check if not an edge beam
            if self.node_spec[ele_nodes[1]]["z_group"] != 0 or self.node_spec[
                ele_nodes[1]
            ]["z_group"] != len(self.noz):
                local_offset = [
                    0,
                    -self.centroid_dist_y,
                    0,
                ]  # shift beam element by global y

        # if element is a transverse member, calculate local offset based on member orientation
        elif (
            self.node_spec[ele_nodes[1]]["x_group"]
            == self.node_spec[ele_nodes[0]]["x_group"]
        ):
            # calculate local offset
            offset_z = self.offset_z_dist  # z == cos
            offset_x = (
                (node_i[0] + node_j[0]) / (node_i[2] + node_j[2]) * self.offset_z_dist
            )
            local_offset = [offset_x, self.y_elevation, offset_z]
        if local_offset:
            if (
                find_min_x_dist(
                    [[a - b for a, b in zip(node_i, local_offset)]], [node_j]
                ).tolist()[0][0]
                < def_l
            ):

                global_offset_i = [a - b for a, b in zip(node_i, local_offset)]
                global_offset_j = [a + b for a, b in zip(node_j, local_offset)]
            else:  # reciprocal , node i has to minus local offset

                global_offset_i = [a + b for a, b in zip(node_i, local_offset)]
                global_offset_j = [a - b for a, b in zip(node_j, local_offset)]
            global_offset = [global_offset_i, global_offset_j]
        vxz = [np.round(num, decimals=self.decimal_lim) for num in vxz]
        tag_value = self.transform_dict.setdefault(
            repr(vxz) + "|" + repr(global_offset), self.transform_counter + 1
        )
        if tag_value > self.transform_counter:
            self.transform_counter = tag_value
        return tag_value


class ShellLinkMesh(Mesh):
    """
    Concrete class for Mesh class. This Mesh class is for model type 3: Shell hybrid model with rigid links
    """

    def __init__(
        self,
        long_dim,
        width,
        trans_dim,
        edge_dist_a,
        edge_dist_b,
        num_trans_beam,
        num_long_beam,
        skew_1,
        skew_2,
        ext_to_int_a,
        ext_to_int_b,
        link_type="beam",
        **kwargs
    ):
        """
        Subclass for mesh with offset beam members linked to grillage consisting of shell elements

        :param long_dim:
        :param width:
        :param trans_dim:
        :param edge_dist_a:
        :param edge_dist_b:
        :param num_trans_beam:
        :param num_long_beam:
        :param skew_1:
        :param skew_2:
        :param ext_to_int_a:
        :param ext_to_int_b:
        """
        # instantiate variables specific for shell mesh subclass
        self.long_ele_offset = []
        self.link_str_list = []
        self.link_type = link_type
        self.link_dict = dict()  # key is node tag c, val is list of master node r
        self.offset_node_group_dict = (
            dict()
        )  # key is node tag of offset node, val is beam group
        self.edge_support_nodes = dict()
        # variables for support - the ends of beam offset elements
        self.pinned_node_group = 0
        self.roller_node_group = 1
        # get variables from keyword arguments
        self.y_offset = kwargs.get("offset_beam_y_dist", 0.449)  # Here default values
        self.beam_width = kwargs.get("beam_width", 0.445)  # Here default values
        self.max_mesh_size_z = kwargs.get("max_mesh_size_z", 1)  # Here default values
        self.max_mesh_size_x = kwargs.get("max_mesh_size_x", 1)  # Here default values
        # variables to store assignment counters after meshing of main model plane grid y = 0
        self.x_grid_to_x_dict = (
            dict()
        )  # key is x value (m), value is x grid number (for offset nodes)
        self.z_grid_to_z_dict = (
            dict()
        )  # key is z value (m), value is z grid number (for offset nodes)

        # use meshing procedure of base mesh class
        super().__init__(
            long_dim,
            width,
            trans_dim,
            edge_dist_a,
            edge_dist_b,
            num_trans_beam,
            num_long_beam,
            skew_1,
            skew_2,
            ext_to_int_a,
            ext_to_int_b,
            **kwargs
        )

        # meshing procedure to create beam offset element and tie it with rigid links to master nodes of model plane y=0
        self._create_offset_beam_element()

        # overwrite procedure to identify
        self._identify_common_z_group()
        # base class stores node groups as self.model_plane_z_groups, here for offset elements store as new variable
        self.offset_z_groups = [
            a
            for a in list(self.z_group_to_ele.keys())
            if a not in self.model_plane_z_groups
        ]

    # -----------------------------------------------------------------------------------------------------------------
    # Functions which are overwritten of that from base class to for specific shell type model
    def create_control_points(self, **kwargs):

        return ShellEdgeControlLine(**kwargs)

    # add groupings of offset beam elements

    # ----------------------------------------------------------------------------------------------------------------
    # sub procedures specific to shell meshes
    def _create_offset_beam_element(self):

        # sub procedure function to create beam elements based on offset nodes
        self._create_offset_nodes()

        # create offset elements commands
        for cNode, rNode_list in self.link_dict.items():
            for rNode in rNode_list:
                self._create_link_element(rNode=rNode, cNode=cNode)

        # loop each beam group and create beam elements from these offset nodes
        for beam_group in range(0, len(self.start_edge_line.z_group_master_pair_list)):
            offset_node_tag = [
                k for k, v in self.offset_node_group_dict.items() if v == beam_group
            ]  # key is offset node
            # sort offset node tag based on x longitudinal position
            x_coord_list = [
                self.node_spec[tag]["coordinate"][0] for tag in offset_node_tag
            ]
            sorted_offset_tag = [
                x for _, x in sorted(zip(x_coord_list, offset_node_tag))
            ]
            # store first and last node tag as supports
            self.edge_support_nodes.setdefault(
                sorted_offset_tag[0], self.pinned_node_group
            )  #
            self.edge_support_nodes.setdefault(
                sorted_offset_tag[-1], self.roller_node_group
            )
            # assign long beam element between two nodes
            for ind, node_tag in enumerate(sorted_offset_tag[:-1]):
                n1 = node_tag
                n2 = sorted_offset_tag[ind + 1]
                transf_tag = self._get_geo_transform_tag(ele_nodes=[n1, n2])
                self.long_ele.append(
                    [
                        self.element_counter,
                        n1,
                        n2,
                        beam_group + self.global_z_grid_count,
                        transf_tag,
                    ]
                )
                self.element_counter += 1

            # add to grouping dict data
            self.z_group_to_ele[beam_group + self.global_z_grid_count] = [
                ele
                for ele in self.long_ele
                if ele[3] == beam_group + self.global_z_grid_count
            ]

    def _create_offset_nodes(self):
        # sub procedure function
        x_count = "offset_beam_x"  # proxy
        z_count = "offset_beam_group_z{}"  # proxy
        # get groups of node master pairs
        z_pair = self.start_edge_line.z_group_master_pair_list
        # loop each z pair
        # loop each x group
        for x_group in range(0, self.global_x_grid_count):
            for beam_group, z_pair_group in enumerate(z_pair):
                n1 = [
                    key
                    for key, n in self.node_spec.items()
                    if n["x_group"] == x_group and n["z_group"] == z_pair_group[0]
                ]
                n2 = [
                    key
                    for key, n in self.node_spec.items()
                    if n["x_group"] == x_group and n["z_group"] == z_pair_group[1]
                ]

                if not len(n1) == 1 or not len(n2) == 1:
                    continue
                n1_coord = self.node_spec[n1[0]]["coordinate"]
                n2_coord = self.node_spec[n2[0]]["coordinate"]

                # create offset node
                mid_pt = [(a + b) / 2 for a, b in zip(n1_coord, n2_coord)]
                node_coordinate = [mid_pt[0], mid_pt[1] + self.y_offset, mid_pt[2]]
                self.node_spec.setdefault(
                    self.node_counter,
                    {
                        "tag": self.node_counter,
                        "coordinate": node_coordinate,
                        "x_group": x_count,
                        "z_group": z_count.format(beam_group),
                    },
                )

                # store offset node rigid details
                master_node_list = [n1[0], n2[0]]  # list of list
                self.link_dict.setdefault(self.node_counter, master_node_list)

                # store node - beam group detail
                # beam_group = [ind for ind, i in enumerate([self.node_spec in z for z in z_pair]) if i][0]  # numbering of beam group
                self.offset_node_group_dict.setdefault(
                    self.node_counter, beam_group
                )  # c node is key, group num is val
                self.node_counter += 1

        # generate for edge nodes - only for orthogonal mesh
        if self.orthogonal:
            for edge_group in range(0, self.global_edge_count):
                for beam_group, z_pair_group in enumerate(z_pair):
                    n1 = [
                        key
                        for key, n in self.node_spec.items()
                        if key in self.edge_node_recorder.keys()
                        and n["z_group"] == z_pair_group[0]
                    ]
                    n2 = [
                        key
                        for key, n in self.node_spec.items()
                        if key in self.edge_node_recorder.keys()
                        and n["z_group"] == z_pair_group[1]
                    ]

                    if not len(n1) == 1 or not len(n2) == 1:
                        continue
                    n1_coord = self.node_spec[n1[0]]["coordinate"]
                    n2_coord = self.node_spec[n2[0]]["coordinate"]

                    # create offset node
                    mid_pt = [(a + b) / 2 for a, b in zip(n1_coord, n2_coord)]
                    node_coordinate = [mid_pt[0], mid_pt[1] + self.y_offset, mid_pt[2]]
                    self.node_spec.setdefault(
                        self.node_counter,
                        {
                            "tag": self.node_counter,
                            "coordinate": node_coordinate,
                            "x_group": x_count,
                            "z_group": z_count.format(beam_group),
                        },
                    )

                    # store offset node rigid details
                    master_node_list = [n1[0], n2[0]]  # list of list
                    self.link_dict.setdefault(self.node_counter, master_node_list)

                    # store node - beam group detail
                    # beam_group = [ind for ind, i in enumerate([self.node_spec in z for z in z_pair]) if i][0]  # numbering of beam group
                    self.offset_node_group_dict.setdefault(
                        self.node_counter, beam_group
                    )  # c node is key, group num is val
                    self.node_counter += 1

    def _create_link_element(self, rNode, cNode):
        # sub procedure function
        # user mp constraint object
        # function to create ops rigid link command and store to variable

        link_str = 'ops.rigidLink("{linktype}",{rNodetag},{cNodetag})\n'.format(
            linktype=self.link_type, rNodetag=cNode, cNodetag=rNode
        )

        self.link_str_list.append(link_str)


# -----------------------------------------------------------------------------------------------------------------
# concrete classes for mesh elements



# -*- coding: utf-8 -*-
"""
Load module consist of class/methods which wraps ```OpenSeesPy``` to create and run load analysis. First part of the module
comprise the user interface functions. Following are the classes of various load types, compound load, load case,
moving load and moving load path.
"""

import pprint
from collections import namedtuple
from collections.abc import Iterable
from copy import deepcopy
from typing import Union

from scipy import interpolate

from ospgrillage.mesh import *


def create_load_vertex(**kwargs):
    """
    User interface function to create load vertex of various load types.
    :param kwargs: Keyword arg, see below.

    :keyword:

    * x (`float` or `int`): Value of x coordinate
    * y (`float` or `int`): Value of y coordinate. Default is model plane y = 0
    * z (`float` or `int`): Value of z coordinate
    * p (`float` or `int`): Magnitude of vertical load in y direction.

    :returns: namedTuple LoadPoint(x,y,z,p)

    :except: ValueError if missing one or more keyword arguments.

    """
    x = kwargs.get("x", None)
    y = kwargs.get("y", 0)
    z = kwargs.get("z", None)
    p = kwargs.get("p", None)

    if not any([x is None, y is None, z is None, p is None]):
        return LoadPoint(x, y, z, p)
    else:
        raise ValueError("Missing one or more keyword arguments for x=, y=, z=, or p=")


def create_point(**kwargs):
    """
    User interface function to create a Point namedTuple.
    :param kwargs: See below
    :keyword:

    * x (`float` or `int`): Value of x coordinate
    * y (`float` or `int`): Value of y coordinate. Default is y = 0
    * z (`float` or `int`): Value of z coordinate

    :return: Point(x,y,z) namedTuple
    """
    x = kwargs.get("x", None)
    y = kwargs.get("y", 0)
    z = kwargs.get("z", None)
    return Point(x, y, z)


def create_load_case(**kwargs):
    """
    User interface function to create LoadCase objects. Following this function, users
    are required add loads to LoadCase object via :func:`~ospgrillage.load.LoadCase.add_load`.

    :keyword:
    * name(`str`): Name string of Load case object

    :returns: :class:`~ospgrillage.load.LoadCase` object
    """
    return LoadCase(**kwargs)


def create_compound_load(**kwargs):
    """
    User interface function to create CompoundLoad object. Following this function users are required to
    add loads to object via :func:`~ospgrillage.load.CompoundLoad.add_load`.

    :keyword:
    name (`str`): Name string of compound load

    :returns: :class:`~ospgrillage.load.CompoundLoad`
    """
    return CompoundLoad(**kwargs)


def create_load(**kwargs):
    """
    User interface function to create load types.

    :keyword:

    * type(`str`): type of load. Choose either ["point","line","patch","nodal"]
    * point# (`LoadPoint` namedTuple): LoadPoint for load type in global coordinate. Note different load type requires a
    different minimum LoadPoint.
    * local_load_point_# (`LoadPoint` namedTuple): LoadPoint for load type in local coordinate. Note different load type
    requires a different minimum LoadPoint.

    :return: PointLoad, LineLoading, PatchLoading, or NodalForces
    """
    type = kwargs.get("loadtype", None)
    if type == "point":
        return PointLoad(**kwargs)
    elif type == "line":
        return LineLoading(**kwargs)
    elif type == "patch":
        return PatchLoading(**kwargs)
    elif type == "nodal":
        fx = kwargs.get("Fx", 0)
        fy = kwargs.get("Fy", 0)
        fz = kwargs.get("Fz", 0)
        mx = kwargs.get("Mx", 0)
        my = kwargs.get("My", 0)
        mz = kwargs.get("Mz", 0)
        tag = kwargs.get("node_tag", None)
        name = kwargs.get("name", None)
        if any(
            [fx is None, fy is None, fz is None, mx is None, my is None, mz is None]
        ):
            raise ValueError(
                "Missing arguments for nodal force definition : Hint check if all required keywords are given"
            )
        force = NodeForces(fx, fy, fz, mx, my, mz)
        return NodalLoad(name=name, node_tag=tag, node_force=force)
    else:
        raise TypeError(
            "load type not specified. hint: specify kwarg type= for create_load()"
        )


def create_moving_load(**kwargs):
    """
    User interface function to create Moving Load object. Following this function, users are required to:

    *. Set a common path to object via :func:`~ospgrillage.load.MovingLoad.set_path`
    *. Add loads to object via :func:`~ospgrillage.load.MovingLoad.add_load`

    :return: :class:`~ospgrillage.load.MovingLoad` object

    :keyword:

    * **common_path**(`Path`): Path object for all load groups added to the Moving load object to traverse
    * **global_increment**(`float` or `int`): Number of increments to discretize Path object. This keyword is only used in
      advance usage where Moving Load contains multiple load groups each with unique path objects.

    """
    return MovingLoad(**kwargs)


def create_moving_path(**kwargs):
    """
    User interface function to create Path object for moving load.

    :keyword:

    * start_point (`Point`): Start point of path
    * end_point (`Point`): End point of path
    * increments (`int`): Increment of path steps. Default is 50
    * mid_point (`Point`): Default = None

    :returns: :class:`~ospgrillage.load.Path` object
    """
    return Path(**kwargs)


# named tuple definition
LoadPoint = namedtuple("Point", ["x", "y", "z", "p"])
NodeForces = namedtuple("node_forces", ["Fx", "Fy", "Fz", "Mx", "My", "Mz"])
Line = namedtuple("line", ["m", "c", "phi"])


# ----------------------------------------------------------------------------------------------------------------
# Loading classes
# ---------------------------------------------------------------------------------------------------------------
class Loads:
    """
    Base class for Point, Line , and Patch loads
    """

    load_point_1: LoadPoint
    load_point_2: LoadPoint
    load_point_3: LoadPoint
    load_point_4: LoadPoint
    load_point_5: LoadPoint
    load_point_6: LoadPoint
    load_point_7: LoadPoint
    load_point_8: LoadPoint

    def __init__(self, **kwargs):
        """

        :param name: Name of load
        :param Fx: Axis force in x axis
        :param Fy: Axis force in y axis
        :param Fz: Axis force in z axis
        :param Mx: Moment about x axis
        :param My: Moment about y axis
        :param Mz: Moment about z axis
        :param kwargs: see below

        :keyword:

        * **point1**, **point2**, ..., **point8** : (LoadPoint namedTuple) coordinate points with force magnitude describing the load type
        * **localpoint1**, **localpoint2**, ..., **localpoint8**: (LoadPoint namedTuple) local coordinate points with force magnitude describing the load type


        """
        #
        self.name = kwargs.get("name", None)

        # Initialise dict for key load points of line UDL and patch load definitions
        self.load_point_data = dict()
        # parse namedtuple of global coordinates
        self.load_point_1 = kwargs.get("point1", None)
        self.load_point_2 = kwargs.get("point2", None)
        self.load_point_3 = kwargs.get("point3", None)
        self.load_point_4 = kwargs.get("point4", None)
        self.load_point_5 = kwargs.get("point5", None)
        self.load_point_6 = kwargs.get("point6", None)
        self.load_point_7 = kwargs.get("point7", None)
        self.load_point_8 = kwargs.get("point8", None)

        # shape function
        self.shape_function = kwargs.get("shape_function", "linear")
        # check if user skipped point 1 and defined point1 as point 2 instead
        if all([self.load_point_1 is None, self.load_point_2 is not None]):
            raise Exception("Load point 1 is not defined")

        # list of load points tuple
        self.point_list = [
            self.load_point_1,
            self.load_point_2,
            self.load_point_3,
            self.load_point_4,
            self.load_point_5,
            self.load_point_6,
            self.load_point_7,
            self.load_point_8,
        ]
        # self.local_point_list = [self.local_load_point_1, self.local_load_point_2, self.local_load_point_3,
        #                          self.local_load_point_4, self.local_load_point_5, self.local_load_point_6,
        #                          self.local_load_point_7, self.local_load_point_8]
        # var for compound load group (handled by LoadCase class when creating compound groups)
        self.compound_dist_x = 0  # local coordinate system
        self.compound_dist_z = 0  # local coordinate system
        self.ref_point = None  # local coordinate system
        self.compound_group = None  # group number access by LoadCase class to move load group if any path is defined
        # spec dict
        self.spec = dict(
            name=self.name, global_points=self.point_list, ref_point=self.ref_point
        )  # dict {node number: {Fx:val, Fy:val, Fz:val, Mx:val, My:val, Mz:val}}
        self.load_counter = 0  # counter for compound load

    # function called by Moving load module to move the load group
    def move_load(self, ref_point: Point):
        """
        Function to move each load point of load type by a reference coordinate. This function is handled by OpsGrillage
        :param ref_point: coordinate to be moved
        :type ref_point: namedTuple Point(x,y,z)
        :return: increment each load point by +x, +y, +z where (x,y,z) is the coordinate prescribed by ref_point
        """
        if any(self.point_list):
            self.load_point_1 = (
                self.load_point_1._replace(
                    x=self.load_point_1.x + ref_point.x,
                    z=self.load_point_1.z + ref_point.z,
                )
                if self.load_point_1 is not None
                else self.load_point_1
            )
            self.load_point_2 = (
                self.load_point_2._replace(
                    x=self.load_point_2.x + ref_point.x,
                    z=self.load_point_2.z + ref_point.z,
                )
                if self.load_point_2 is not None
                else self.load_point_2
            )
            self.load_point_3 = (
                self.load_point_3._replace(
                    x=self.load_point_3.x + ref_point.x,
                    z=self.load_point_3.z + ref_point.z,
                )
                if self.load_point_3 is not None
                else self.load_point_3
            )
            self.load_point_4 = (
                self.load_point_4._replace(
                    x=self.load_point_4.x + ref_point.x,
                    z=self.load_point_4.z + ref_point.z,
                )
                if self.load_point_4 is not None
                else self.load_point_4
            )
            self.load_point_5 = (
                self.load_point_5._replace(
                    x=self.load_point_5.x + ref_point.x,
                    z=self.load_point_5.z + ref_point.z,
                )
                if self.load_point_5 is not None
                else self.load_point_5
            )
            self.load_point_6 = (
                self.load_point_6._replace(
                    x=self.load_point_6.x + ref_point.x,
                    z=self.load_point_6.z + ref_point.z,
                )
                if self.load_point_6 is not None
                else self.load_point_6
            )
            self.load_point_7 = (
                self.load_point_7._replace(
                    x=self.load_point_7.x + ref_point.x,
                    z=self.load_point_7.z + ref_point.z,
                )
                if self.load_point_7 is not None
                else self.load_point_7
            )
            self.load_point_8 = (
                self.load_point_8._replace(
                    x=self.load_point_8.x + ref_point.x,
                    z=self.load_point_8.z + ref_point.z,
                )
                if self.load_point_8 is not None
                else self.load_point_8
            )

    def apply_load_factor(self, factor=1):
        """
        Function to apply load factor to each load point's p value (vertical P force)
        """
        self.load_point_1 = (
            self.load_point_1._replace(p=factor * self.load_point_1.p)
            if self.load_point_1 is not None
            else self.load_point_1
        )
        self.load_point_2 = (
            self.load_point_2._replace(p=factor * self.load_point_2.p)
            if self.load_point_2 is not None
            else self.load_point_2
        )
        self.load_point_3 = (
            self.load_point_3._replace(p=factor * self.load_point_3.p)
            if self.load_point_3 is not None
            else self.load_point_3
        )
        self.load_point_4 = (
            self.load_point_4._replace(p=factor * self.load_point_4.p)
            if self.load_point_4 is not None
            else self.load_point_4
        )
        self.load_point_5 = (
            self.load_point_5._replace(p=factor * self.load_point_5.p)
            if self.load_point_5 is not None
            else self.load_point_5
        )
        self.load_point_6 = (
            self.load_point_6._replace(p=factor * self.load_point_6.p)
            if self.load_point_6 is not None
            else self.load_point_6
        )
        self.load_point_7 = (
            self.load_point_7._replace(p=factor * self.load_point_7.p)
            if self.load_point_7 is not None
            else self.load_point_7
        )
        self.load_point_8 = (
            self.load_point_8._replace(p=factor * self.load_point_8.p)
            if self.load_point_8 is not None
            else self.load_point_8
        )

    def get_magnitude(self):
        # TODO
        magnitude = []
        for load_point in self.point_list:
            magnitude.append(load_point.p)

        return magnitude

    def __str__(self):
        return "Load object {} \n".format(self.name) + pprint.pformat(self.spec)


class NodalLoad(Loads):
    """
    Main class for nodal load. Derived from Loads base class
    """

    def __init__(self, node_tag, node_force, name=None):
        """
        Nodal load takes a node tag and namedtuple NodalForce(Fx,Fy,Fz,Mx,My,Mz) as input.

        :param name: Name of load
        :type name: str
        :param node_tag: Node tag of grillage model for nodal load to be applied
        :type node_tag: int
        :param node_force: Named tuple of node forces
        :type node_force: NodalForces(Fx,Fy,Fz,Mx,My,Mz)
        """
        super().__init__(name=name, node_tag=node_tag)
        self.Fx = node_force.Fx
        self.Fy = node_force.Fy
        self.Fz = node_force.Fz
        self.Mx = node_force.Mx
        self.My = node_force.My
        self.Mz = node_force.Mz
        self.node_tag = node_tag
        if not isinstance(node_tag, Iterable):
            node_list = [node_tag]
        else:
            node_list = node_tag
        for nodes in node_list:
            self.spec[nodes] = {
                "Fx": self.Fx,
                "Fy": self.Fy,
                "Fz": self.Fz,
                "Mx": self.Mx,
                "My": self.My,
                "Mz": self.Mz,
            }

    def get_nodal_load_str(self):
        """
        Function to return ops.load() command for nodal load.
        """
        # get str for ops.load() function.
        load_value = [self.Fx, self.Fy, self.Fz, self.Mx, self.My, self.Mz]
        load_str = "ops.load({pt}, *{val})\n".format(pt=self.node_tag, val=load_value)
        return load_str


class PointLoad(Loads):
    """
    Class for Point loads. Derived from based :py:class:`Loads` class
    """

    def __init__(self, **kwargs):
        """

        :param name:
        :param kwargs:
        """
        super().__init__(**kwargs)


class LineLoading(Loads):
    """
    Class for line loading. Derived from based Loads class
    """

    def __init__(self, **kwargs):
        """

        :param name:
        :param kwargs:
        """
        super().__init__(**kwargs)

        self.long_beam_ele_load_flag = kwargs.get("long_beam_element_load", False)
        self.trans_beam_ele_load_flag = kwargs.get("trans_beam_element_load", False)

        # if local coordinate is defined, create parameters based on local coordinates. later, if add_global_coord()
        # is called, set their load_point_
        # if not any(self.point_list) and any(self.local_point_list):
        #     if self.local_load_point_3 is not None:  # curve
        #         # findCircle assumes model plane is y = 0, ignores y input, y in this case is a 2D view of x z plane
        #         self.d = findCircle(x1=self.local_load_point_1.x, y1=self.local_load_point_1.z,
        #                             x2=self.local_load_point_2.x, y2=self.local_load_point_2.z,
        #                             x3=self.local_load_point_3.x, y3=self.local_load_point_3.z)
        #         # return a function variable
        #         self.line_end_point = self.local_load_point_3
        #     else:  # straight line with 2 points
        #         self.m_local, self.phi_local = get_slope(
        #             [self.local_load_point_1.x, self.local_load_point_1.y, self.local_load_point_1.z],
        #             [self.local_load_point_2.x, self.local_load_point_2.y, self.local_load_point_2.z])
        #         self.c_local = get_y_intcp(m=self.m_local, x=self.local_load_point_1.x, y=self.local_load_point_1.z)
        #         self.angle_local = np.arctan(self.m_local) if self.m_local is not None else np.pi / 2  # in radian
        #         self.line_end_point = self.local_load_point_2
        #         # namedTuple Line
        #         self.line_equation = Line(self.m_local, self.c_local, self.phi_local)

        # elif not any(self.local_point_list) and any(self.point_list):
        # if three points are defined, set line as curved circular line with point 2 (x2,y2,z2) in the centre of
        # curve
        if self.load_point_3 is not None:  # curve
            # findCircle assumes model plane is y = 0, ignores y input, y in this case is a 2D view of x z plane
            self.d = find_circle(
                x1=self.load_point_1.x,
                y1=self.load_point_1.z,
                x2=self.load_point_2.x,
                y2=self.load_point_2.z,
                x3=self.load_point_3.x,
                y3=self.load_point_3.z,
            )
            # return a function variable
            self.line_end_point = self.load_point_3
        else:  # straight line with 2 points
            self.m, self.phi = get_slope(
                [self.load_point_1.x, self.load_point_1.y, self.load_point_1.z],
                [self.load_point_2.x, self.load_point_2.y, self.load_point_2.z],
            )
            self.c = get_y_intcp(m=self.m, x=self.load_point_1.x, y=self.load_point_1.z)
            self.angle = (
                np.arctan(self.m) if self.m is not None else np.pi / 2
            )  # in radian
            self.line_end_point = self.load_point_2
            # namedTuple Line
            self.line_equation = Line(self.m, self.c, self.phi)
        # else:
        #     raise ValueError("Invalid load points for line load {}".format(self.name))

    def interpolate_udl_magnitude(self, point_coordinate):
        #   """
        #   Function to interpolate magnitude of load point between two load points in a line segment.
        #
        #   Example illustration: Function returns p @ [x y z]
        #
        #   p(loadpoint1)_____p(x=,y,z)_______ p(loadpoint2)
        #   ||||||||||||||||||||||||||||||||||||||||||||||      Line loading
        #   ||||||||||||||||||||||||||||||||||||||||||||||
        # __________________________________________________________
        #
        #   :param point_coordinate: coordinate list [x,y,z]
        #   :type point_coordinate: list
        #   :return: point force (udl) magnitude at coordinate
        #   """
        # input: point_coordinate list of [x,y,z]
        pp = None
        # check if line is straight or curve
        if self.load_point_3 is None:  # straight line

            # x[0],z[0] and p[0] shall be reference point for interpolate
            xp = point_coordinate[0]
            yp = point_coordinate[0]  # not used but generated here
            zp = point_coordinate[0]

            # use parametric equation of line in 3D
            v = [
                self.load_point_2.x - self.load_point_1.x,
                self.load_point_2.p - self.load_point_1.p,
                self.load_point_2.z - self.load_point_1.z,
            ]
            if v[0] == 0 and self.load_point_2.x == self.load_point_1.x:
                pp = (zp - self.load_point_1.z) / v[2] * v[1] + self.load_point_1.p
            else:
                pp = (xp - self.load_point_1.x) / v[0] * v[1] + self.load_point_1.p

        elif self.load_point_3 is not None:  # curve
            # TODO for curved line load
            pass
        return pp

    def get_point_given_distance(self, xbar, point_coordinate):
        # """
        # Function to return
        # :param xbar: distance
        # :type xbar: float
        # :param point_coordinate: coordinates list [x,y,z]
        # :type point_coordinate: list
        # :return new_point: coordinate list [x,y,z] shifted by distance
        # :type new_point: list
        # """
        # function to return centroid of line load given reference point coordinate (point2) and xbar calculated based
        # on
        z_dis = xbar * np.sin(self.angle)
        x_dis = xbar * np.cos(self.angle)
        # y dis = 0 due to model plane
        new_point = [
            point_coordinate[0] - x_dis,
            point_coordinate[1],
            point_coordinate[2] - z_dis,
        ]
        return new_point

    def get_line_segment_given_x(self, x):
        # """
        # Function to return straight line equation for line segment (in OpsGrillage case, segment bounded by grid) given x point
        # :param x: value of x input for line equation
        # :type x: float
        # :return: solution of line equation (i.e. y = mx + c)
        # """
        if self.line_equation.m is None:  # if vertical line
            pass
        else:
            if (
                self.load_point_1.x <= x <= self.line_end_point.x
                or self.load_point_1.x >= x >= self.line_end_point.x
            ):
                return line_func(self.line_equation.m, self.line_equation.c, x)

    def get_line_segment_given_z(self, z):
        if self.line_equation.m is None:  # if vertical line
            if (
                self.load_point_1.z <= z <= self.line_end_point.z
                or self.load_point_1.z >= z >= self.line_end_point.z
            ):
                return self.load_point_1.x
        else:
            if (
                self.load_point_1.z <= z <= self.line_end_point.z
                or self.load_point_1.z >= z >= self.line_end_point.z
            ):
                return inv_line_func(self.line_equation.m, self.line_equation.c, z)


class PatchLoading(Loads):
    """
    Main class for Patch loads. Derived from base Loads class.

    Patch load can take up to 8 load points. By default requires at least 4 load point for patch (quadrilateral)
    """

    def __init__(self, **kwargs):
        """

        :param name:
        :param kwargs:
        """
        super().__init__(**kwargs)
        if not all(v is None for v in self.point_list):
            a, _ = sort_vertices(
                [
                    self.load_point_2,
                    self.load_point_3,
                    self.load_point_1,
                    self.load_point_4,
                ]
            )
            match = [
                self.load_point_1,
                self.load_point_2,
                self.load_point_3,
                self.load_point_4,
            ]
            # if len(self.point_list) < 4:
            #     raise ValueError("invalid number of vertices. Hint:  either 4 or 8 is accepted for patch")

        else:
            raise ValueError(
                "vertices missing.hint: patch load must have either 4  or 8 vertices  "
            )
        if a != match:
            raise Exception(
                "vertices of patch load gives invalid patch layout: hint: make sure vertices are in counter"
                "clockwise order with first point (load_point_1) being the bottom left vertice of the "
                "patch load"
            )
        # get patch min dimension
        self.patch_min_dim = None  # instantiate
        # procedure to define lines of patch
        self._define_patch_edge_lines()

    def _define_patch_edge_lines(self):
        # create each line
        self.line_1 = LineLoading(point1=self.load_point_1, point2=self.load_point_2)
        self.line_2 = LineLoading(point1=self.load_point_2, point2=self.load_point_3)
        self.line_3 = LineLoading(point1=self.load_point_3, point2=self.load_point_4)

        # if only four point is define , patch load is a four point straight line quadrilateral
        if self.load_point_5 is None:
            # create fourth line
            self.line_4 = LineLoading(
                point1=self.load_point_4, point2=self.load_point_1
            )

            # create equation of plane from four straight lines

            # create interpolate object f
            p = np.array(
                [
                    [self.load_point_1.p, self.load_point_2.p],
                    [self.load_point_3.p, self.load_point_4.p],
                ]
            )
            x = np.array(
                [
                    [self.load_point_1.x, self.load_point_2.x],
                    [self.load_point_3.x, self.load_point_4.x],
                ]
            )
            z = np.array(
                [
                    [self.load_point_1.z, self.load_point_2.z],
                    [self.load_point_3.z, self.load_point_4.z],
                ]
            )

            # create function to get interpolation of p
            self.patch_mag_interpolate = interpolate.interp2d(x, z, p)
            mod_list = [ls for ls in self.point_list if ls is not None]
            self.patch_min_dim = min(
                [
                    get_distance(p1, p2)
                    for (p1, p2) in zip(mod_list, mod_list[1:] + [mod_list[0]])
                    if all([p1 is not None, p2 is not None])
                ]
            )
        elif self.load_point_4 is None:
            # update line 3
            self.line_3 = LineLoading(
                point1=self.load_point_3, point2=self.load_point_1
            )

            # TODO create equation of plane from 3 points
            # create interpolate object f
            p = np.array(
                [
                    [self.load_point_1.p, self.load_point_2.p],
                    [self.load_point_3.p, self.load_point_4.p],
                ]
            )
            x = np.array(
                [
                    [self.load_point_1.x, self.load_point_2.x],
                    [self.load_point_3.x, self.load_point_4.x],
                ]
            )
            z = np.array(
                [
                    [self.load_point_1.z, self.load_point_2.z],
                    [self.load_point_3.z, self.load_point_4.z],
                ]
            )

            # create function to get interpolation of p
            self.patch_mag_interpolate = interpolate.interp2d(x, z, p)
            mod_list = [ls for ls in self.point_list if ls is not None]
            self.patch_min_dim = min(
                [
                    get_distance(p1, p2)
                    for (p1, p2) in zip(mod_list, mod_list[1:] + [mod_list[0]])
                    if all([p1 is not None, p2 is not None])
                ]
            )
        elif self.load_point_8 is not None:
            # TODO
            # point 1 2 3
            # point 3 4 5
            # point 5 6 7
            # point 7 8 1
            pass

        else:
            raise ValueError(
                "Patch load points for patch load {} not valid".format(self.name)
            )


# ---------------------------------------------------------------------------------------------------------------
class CompoundLoad:
    """
    Main class for Compound load definition.

    When a Load object is pass as an input, CompoundLoad treats the initial positions (load_points) of the Load classes
    as local coordinates. Then CompoundLoad sets the loads "in-place" of the local coordinate. If class input local_coord
    is given, CompoundLoad replaces the coordinates of the initial load points (retaining the magnitude of load points)

    When set_global_coord() function is called, CompoundLoad sets the input global coordinate as the new centroid of the
    compounded load groups. This is done by shifting each local coordinate load point in all load groups under
    CompoundLoad by x (global) and z (global).

    Here are a few relationships between CompoundLoad and other classes

    * CompoundLoad object can have Loads class and its inheritances (Line, Point, Patch)
    * CompoundLoad handles functions of Load classes (e.g. move_load)

    """

    def __init__(self, name):
        self.name = name
        self.compound_load_obj_list = []
        self.local_coord_list = []
        self.centroid = Point(0, 0, 0)  # named tuple Point
        self.global_coord = self.centroid

    def add_load(self, load_obj: Loads):
        """
        Function to add load object to compound load group. If a local_coord parameter is given, this new local_coord overwrites the coordinates (either local or global) of the load object.

        ..note:
            If load object is defined using local coordinate and local_coord is None, its default local coord precedes.

        :param load_obj: Load object
        :type load_obj: PointLoad,LineLoading,PatchLoading
        :param local_coord: Local coordinate of load object
        :type local_coord: Point namedTuple

        """
        # update the load obj to be part of compound load by first
        # shifting all load points relative to centroid of defined load class
        # then shifting centroid and load_points relative to A Local Coordinate system
        load_obj_copy = deepcopy(load_obj)
        # check if input load object is valid (local vs global coordinate) system.
        # If local load + local_coord input, raise ValueError
        # if local_coord is not None and any(load_obj_copy.local_point_list):
        #     raise ValueError("{} was defined in local coordinate space. However a `load_coord=` parameter"
        #                      "exist for this load when creating compound load "
        #                      "Hint: Loads defined in local coordinate space does not "
        #                      "need another "
        #                      "`local_coord=` parameter".format(load_obj_copy.name))
        # if global load + local_coord input, raise Value error
        # if local_coord is not None and not any(load_obj_copy.local_point_list):
        #    raise ValueError("{} was defined in global coordinate. However a local_coord= parameter exist for this load"
        #                     ".Hint: Load defined in global coordinate space does not need `local_coord=` "
        #                     "parameter".format(load_obj_copy.name))
        #    pass
        # inputs for compound load are valid, proceed to set compound load
        # if local coord is passed in,
        # if local_coord is not None:  # else points defined as global already
        #     load_obj_copy.form_compound_load(compound_dist_x=local_coord.x, compound_dist_z=local_coord.z)
        #     # then shift load obj relative to global coord (this is the coord of the model) default is 0,0,0 if not set
        #     # by user
        #     load_obj_copy.move_load(self.global_coord)

        self.compound_load_obj_list.append(
            load_obj_copy
        )  # after update, append to list
        # self.local_coord_list.append(local_coord)

    def set_global_coord(self, global_coord: Point):
        """
        Function to set global coordinate of the compound load with respect to global coordinate system of grillage model.
        The global coordinate is set to all local load points (i.e. it adds the global coord x y z to each local coord)

        :param global_coord: Value of x y z (global coord) to offset the basic coordinate system
        :type global_coord: Point namedTuple
        """
        # overwrite global coordinate
        if global_coord != self.global_coord:
            self.global_coord = global_coord
            # loop each load type in compound load and shift by global_coord
            append_load_list = []
            # once overwritten, update loadpoints in each load obj of compound load by global_coord
            for loads in self.compound_load_obj_list:
                loads.move_load(
                    global_coord
                )  # shift load objs using Load class method move_load()
                append_load_list.append(
                    loads
                )  # append loads which have been moved to new list and
            self.compound_load_obj_list = (
                append_load_list  # overwrite it to class variable
            )


# ---------------------------------------------------------------------------------------------------------------
class LoadCase:
    """
    Main class for load cases. Each load case holds information about:
    #. Load object types (point line patch or combination i.e. compound load)
    #. load case ops load command line - this is generated by ops-grillage class method (add_load_case()) and updated to LoadCase class object.
    #. Load factor - all load objects within a Load case are link to one same load factor


    Here are a few relationships between LoadCase and other classes

    * MovingLoad class creates a series of load case representing the movement of load objects in each load case.
    * Load combination takes in several LoadCase class instance with varied load factors into a single analysis
    * Analysis class handles the ops. commands required for
    * OpsGrillage class takes in load case and updates the variable 'load_command_list' after distributing loads within the LoadCase class to nodes/elements of the Mesh in OpsGrillage.
    * LoadCase class can have Load objects or CompoundLoad class object

    """

    def __init__(self, name):
        """
        :param name: str of load case name
        """
        self.name = name
        self.load_groups = []
        self.position = None
        # preset load factor for
        self.load_command_list = []

    def add_load(self, load_obj: Union[Loads, CompoundLoad], **kwargs):
        """
        Function to add load objects to LoadCase

        :param load_obj: Load or Compound load object
        :param kwargs: keyword arguments
        :keyword:

        * global_coord_of_load_obj (`Point` namedTuple): if load objects are defined in local coordinate, this parameter
          is required to set the origin of local coordinate of load groups onto the global coordinate of the grillage

        """
        load_dict = dict()
        load_dict.setdefault(
            "load", deepcopy(load_obj)
        )  # create copy of object instance
        # check if load_obj's load points are local points, if True, check if kwargs global coord is provided
        global_coord_of_load_obj: Point = kwargs.get("global_coord_of_load_obj", None)

        load_factor = kwargs.get("load_factor", 1)
        load_dict.setdefault("factor", load_factor)
        self.load_groups.append(load_dict)

    # function for if load groups are to change its ref position due to movement / traversing loads
    # warning : this function is only to be handled by MovingLoad class
    def move_load_group(self, ref_point: Point):
        self.position = ref_point
        for load_dict in self.load_groups:
            load_obj = load_dict.get("load")
            if isinstance(load_obj, CompoundLoad):
                for ind_load_obj in load_obj.compound_load_obj_list:
                    ind_load_obj.move_load(self.position)
            else:
                load_obj.move_load(self.position)


# ---------------------------------------------------------------------------------------------------------------
class MovingLoad:
    """
    Main class of moving load case. MovingLoad class parses and creates multiple loadcase object corresponding to
    traversing the input load groups - be it compound or single. Moving load is able to set various path (defined by
    Path class object) to individual load groups.

    """

    def __init__(self, name, **kwargs):
        """

        :param name: Name string of moving load
        :keyword:
        * common_path (`Path`): Path object specifying the common path for all loads defined in moving load to traverse
        * global_increment(`int`): Number of increments to discretize Path object.

        .. note::
            global_increment argument is used in advance moving load analysis, where a moving load object attributes
            each assigned load type with a corresponding unique Path object (which can differ between different paths).
        """
        self.name = name
        self.load_list = []
        self.load_case_dict_list = []  # Variable to access
        self.moving_load_case = (
            []
        )  # Variable recording all created load case for all load group's respective path
        self.static_load_case = []
        # get kwargs
        self.common_path = kwargs.get("common_path", None)
        self.global_increment = kwargs.get("global_increment", None)  # for advance use
        self.parse = False  # flag for if query option is available
        self.incremental_name = None  # init variable of query method

    def set_path(self, path_obj):
        """
        Function to assign/modify the common path variable with a new Path object.
        All loads added later to Moving load object will traverse the same common path object.

        :param path_obj: Path object to specify common path variable.
        :type path_obj: Path

        """
        if self.global_increment is not None:
            raise ValueError(
                "Option for Basic use - common path defined for all added loads in MovingLoad however"
                "a global increment parameter was defined. Hint: Remove global_increment= on creating"
                "moving load object"
            )
        # else, valid input for setting a basic moving load - proceed setting common path variable
        self.common_path = path_obj

    def add_load(self, load_obj: Union[Loads, CompoundLoad], path_obj=None):
        """
        Function to set a load type (Loads class object) with its path (Path class object). Function accepts compound load (Compound load class) as a load input, which in turn sets the path object to all loads within the compound
        load group.

        :param load_obj: Loads class object , or Compound load object
        :param path_obj: Path class object - this is for advance use, where users specify unique path object for each load within the moving load object.

        """
        # if no path object is added, set empty list to path_obj. The load group will be treated as a static load
        # present throughout the movement of other load groups (added to the series of moving load case)

        load_pair_path = dict()
        load_pair_path.setdefault("load", load_obj)
        # check if basic moving load case
        if self.common_path and self.global_increment is None:
            load_pair_path.setdefault(
                "path", self.common_path.get_path_points()
            )  # .get_path_points() class method of Path class
        elif (
            self.global_increment
        ):  # advance use - where each object has individual path
            load_pair_path.setdefault(
                "path", path_obj.get_custom_path_points(self.global_increment)
            )
        else:  # error, no global statement was provided,
            raise ValueError(
                "No set_path() for moving load {}: Hint run set_path() before add_loads()".format(
                    self.name
                )
            )
        self.load_case_dict_list.append(load_pair_path)

    # function to create incremental load cases for each step of the moving loads. Function handled by OspGrillage
    def parse_moving_load_cases(self):

        # loop through all load-path pairs and identify static loads
        for load_pair_dict in self.load_case_dict_list:
            if not load_pair_dict["path"]:  # empty path, load is static
                self.static_load_case.append(load_pair_dict["load"])

        # create load case obj for each step in the move
        for load_pair_dict in self.load_case_dict_list:
            path_list = load_pair_dict["path"]  # extract path_list of load object
            load_obj = load_pair_dict["load"]
            # loop to create a load case for each increment of the path obj
            load_case_list = []
            for steps in path_list:
                load_step_lc = LoadCase(
                    name="{} at global position [{:.2f},{:.2f},{:.2f}]".format(
                        self.name, steps[0], steps[1], steps[2]
                    )
                )  # _lc in name stands for load case
                load_obj_copy = deepcopy(
                    load_obj
                )  # Use deepcopy module to copy instance of load
                load_step_lc.add_load(
                    load_obj_copy
                )  # and add load to newly created load case
                # add entries of static load to load groups
                step_point = Point(
                    steps[0], steps[1], steps[2]
                )  # convert increment position into Point tuple
                load_step_lc.move_load_group(
                    step_point
                )  # increment the load groups by step point
                # static load not used
                for (
                    static_load
                ) in (
                    self.static_load_case
                ):  # add static load portions to each incremental load case
                    static_load_copy = deepcopy(static_load)
                    load_step_lc.add_load(static_load_copy)
                load_case_list.append(load_step_lc)
            self.moving_load_case.append(load_case_list)
            self.parse = True
        return self.moving_load_case

    def query(self, incremental_lc_name, **kwargs):
        """
        Function to query properties of moving load

        :param incremental_lc_name: Name string of load case to query properties
        :type incremental_lc_name: str
        :param kwargs:
        :return:
        """
        # get query options
        if not self.parse:
            raise Exception(
                "Moving load is not yet set to a grillage model - no information ready for query. hint"
                "add moving load to a grillage model via add_load_case()"
            )
        # specific incremental load case name
        self.incremental_name = incremental_lc_name
        # specific load group name
        load_group_name = kwargs.get("load_group_name", [])
        index = kwargs.get("index", 0)
        option = kwargs.get("option", "position")

        # instantiate variables
        query_load_case = []
        # get incremental load cases based on incremental names provided by users
        for ml in self.moving_load_case:
            for incremental_lc in ml:
                if self.incremental_name == incremental_lc.name:
                    query_load_case = incremental_lc
                    break
            break
        # get load groups within moving load and its respective path
        if load_group_name:
            selected_load_groups = [
                load
                for load in query_load_case.load_groups
                if load["load"].name in load_group_name
            ]
            selected_lc_list = [
                a for a in self.load_case_dict_list if a["load"].name in load_group_name
            ]
        else:  # select all load groups in
            selected_load_groups = query_load_case.load_groups
            selected_lc_list = self.load_case_dict_list

        # TODO add more options* and check if compound or basic
        if option == "position":
            return [a["load"].point_list for a in selected_load_groups]
        elif option == "offset":
            return [
                a["load"].point_list - b
                for (a, b) in zip(selected_load_groups, selected_lc_list)
            ]
        elif option == "original":
            return selected_lc_list
        elif option == "path":
            return [load["path"] for load in selected_lc_list]


class Path:
    """
    Main class to define path of a moving load object
    """

    def __init__(
        self,
        start_point: Point,
        end_point: Point,
        increments=50,
        mid_point: Point = None,
    ):
        self.start_point = start_point
        self.end_point = end_point
        # here create a straight path
        self.path_points_x = np.linspace(start_point.x, end_point.x, increments)
        self.path_points_y = np.linspace(start_point.y, end_point.y, increments)
        self.path_points_z = np.linspace(start_point.z, end_point.z, increments)
        self.path_points_list = []

    def get_path_points(self) -> list:
        self.path_points_list = [
            [x, y, z]
            for (x, y, z) in zip(
                self.path_points_x, self.path_points_y, self.path_points_z
            )
        ]
        return self.path_points_list

    def get_custom_path_points(self, new_increment):
        path_points_x = np.linspace(self.start_point.x, self.end_point.x, new_increment)
        path_points_y = np.linspace(self.start_point.x, self.end_point.x, new_increment)
        path_points_z = np.linspace(self.start_point.x, self.end_point.x, new_increment)
        path_point_list = [
            [x, y, z] for (x, y, z) in zip(path_points_x, path_points_y, path_points_z)
        ]
        return path_point_list


# ---------------------------------------------------------------------------------------------------------------
def create_load_model(**kwargs):
    """
    Function to create a CompoundLoad object of a vehicle load model model
    :return: `LoadModel` object
    """
    return LoadModel(**kwargs)
    pass


class LoadModel:
    """
    Class to handle load model generator. This contains library of load models and creates load model using
    `CompoundLoad` class.

    For users wishing to contribute/add load models, do so herein by adding the load model as a function to
    this class.

    """

    def __init__(self, gap=0, **kwargs):
        """
        Init the class.

        :param gap: Gap between axle (specific to M1600)
        :param kwargs: See below
        """
        self.gap = gap
        self.model_type = kwargs.get("model_type", None)
        self.units = kwargs.get("units", "SI")
        self.origin = kwargs.get("origin", Point(0, 0, 0))
        # create unit dict
        self.unit_dict = {"m": 1, "kN": 1e3}
        # create variables
        self.x_offset = self.origin.x
        self.z_offset = self.origin.z
        # default y offset is zero

        # checks
        if self.origin is None:
            self.x_offset = 0

    def create(self):
        if self.model_type == "M1600":
            return self.create_m1600_vehicle(self.gap)

    def create_m1600_vehicle(self, gap):
        """
        AS5100 Australian load model.

        :param gap: Gap between axle group
        :return: :class:`~ospgrillage.load.CompoundLoad` object of a M1600 vehicle in local coordinate.
        """
        # default SI units
        m = 1
        kN = 1e3
        if self.units == "SI":  # if SI, de-activate converter variables
            ft_convert = 1
            ton_convert = 1
        else:  # Imperial units
            ft_convert = 3.28084
            ton_convert = 0.10036113565668

        axle_dist = 1.25 * m * ft_convert
        left_group_dist = 3.75 * m * ft_convert
        right_group_dist = 5 * m * ft_convert
        wheel_load = 60 * kN * ton_convert
        load_positions_x = [
            0,
            axle_dist,
            axle_dist * 2,
            left_group_dist + axle_dist * 2,
            left_group_dist + axle_dist * 2 + axle_dist,
            left_group_dist + axle_dist * 2 + 2 * axle_dist,
            gap + left_group_dist + axle_dist * 2,
            gap + left_group_dist + axle_dist * 2 + axle_dist,
            gap + left_group_dist + axle_dist * 2 + 2 * axle_dist,
            right_group_dist + gap + left_group_dist + axle_dist * 2,
            right_group_dist + gap + left_group_dist + axle_dist * 2 + axle_dist,
            right_group_dist + gap + left_group_dist + axle_dist * 2 + 2 * axle_dist,
        ]
        load_positions_x = [point + self.x_offset for point in load_positions_x]

        load_positions_z = [-1, 1]
        load_positions_z = [point + self.z_offset for point in load_positions_z]

        M1600_vehicle = create_compound_load(name="M1600 Vehicle")
        for z in load_positions_z:
            for x in load_positions_x:
                vert = create_load_vertex(x=x, z=z, p=wheel_load)
                point = create_load(loadtype="point", name="M1600 point", point1=vert)
                M1600_vehicle.add_load(load_obj=point)

        return M1600_vehicle


# ---------------------------------------------------------------------------------------------------------------
class ShapeFunction:
    """
    Class for shape functions. The role of Shape functions in ospgrillage is to distribute loads to nodes.

    Here developers can add more shape functions to this class by:
    * adding the option in get_shape_function()
    * defining the shape function as a class function

    """

    def __init__(self, option_three_node="triangle_linear", option_four_node="hermite"):
        self.option_three_node = option_three_node
        self.option_four_node = option_four_node

    def get_shape_function(self, option, eta=0, zeta=0):
        if option == "hermite":
            return lambda: self.hermite_shape_function_2d(eta, zeta)
        elif option == "linear":
            return lambda: self.linear_shape_function(eta, zeta)
        elif option == "triangle_linear":
            return lambda: self.linear_triangular

    @staticmethod
    def hermite_shape_function_1d(
        zeta, a
    ):  # using zeta and a as placeholders for normal coor + length of edge element
        # hermite shape functions
        """
        :param zeta: absolute position in x direction
        :param a: absolute position in x direction
        :return: Four terms [N1, N2, N3, N4] of hermite shape function
        .. note::

        """
        N1 = 1 - 3 * zeta ** 2 + 2 * zeta ** 3
        N2 = (zeta - 2 * zeta ** 2 + zeta ** 3) * a
        N3 = 3 * zeta ** 2 - 2 * zeta ** 3
        N4 = (-(zeta ** 2) + zeta ** 3) * a
        return [N1, N2, N3, N4]

    @staticmethod
    def hermite_shape_function_2d(eta, zeta):
        # nodes must be counter clockwise such that n1 = left bottom of relative grid
        # 4  3
        # 1  2
        h1 = 0.25 * (2 - 3 * eta + eta ** 3)
        h2 = 0.25 * (1 - eta - eta ** 2 + eta ** 3)
        h3 = 0.25 * (2 + 3 * eta - eta ** 3)
        h4 = 0.25 * (-1 - eta + eta ** 2 + eta ** 3)
        z1 = 0.25 * (2 - 3 * zeta + zeta ** 3)
        z2 = 0.25 * (1 - zeta - zeta ** 2 + zeta ** 3)
        z3 = 0.25 * (2 + 3 * zeta - zeta ** 3)
        z4 = 0.25 * (-1 - zeta + zeta ** 2 + zeta ** 3)
        Nv = [h1 * z1, h3 * z1, h3 * z3, h1 * z3]
        Nmz = [h2 * z1, h4 * z1, h4 * z3, h2 * z3]
        Nmx = [h1 * z2, h3 * z2, h3 * z4, h1 * z4]
        return Nv, Nmx, Nmz

    @staticmethod
    def linear_shape_function(eta, zeta):
        """
        :param zeta: absolute position in x direction
        :param eta: absolute position in z direction
        :return: Four terms [N1, N2, N3, N4] of Linear shape function
        .. note::
            Further validation needed - trial on different bridge models
        """
        N1 = 0.25 * (1 - eta) * (1 - zeta)
        N2 = 0.25 * (1 + eta) * (1 - zeta)
        N3 = 0.25 * (1 + eta) * (1 + zeta)
        N4 = 0.25 * (1 - eta) * (1 + zeta)
        return [N1, N2, N3, N4]

    @staticmethod
    def linear_triangular(x, z, x1, z1, x2, z2, x3, z3):
        # modelling plane = y plane
        ae = 0.5 * ((x2 * z3 - x3 * z2) + (z2 - z3) * x1 + (x3 - x2) * z1)
        a1 = (x2 * z3 - x3 * z2) / (2 * ae)
        b1 = (z2 - z3) / (2 * ae)
        c1 = (x3 - x2) / (2 * ae)

        a2 = (x3 * z1 - x1 * z3) / (2 * ae)
        b2 = (z3 - z1) / (2 * ae)
        c2 = (x1 - x3) / (2 * ae)

        a3 = (x1 * z2 - x2 * z1) / (2 * ae)
        b3 = (z1 - z2) / (2 * ae)
        c3 = (x2 - x1) / (2 * ae)
        N1 = a1 + b1 * x + c1 * z
        N2 = a2 + b2 * x + c2 * z
        N3 = a3 + b3 * x + c3 * z
        return [N1, N2, N3]
